
class RotorSE_Example2():

	def execute(self):

		# --- Import Modules
		import numpy as np
		import os
		from openmdao.api import IndepVarComp, Component, Group, Problem, Brent, ScipyGMRES, ScipyOptimizer, DumpRecorder
		from rotorse.rotor_aeropower import RotorAeroPower
		from rotorse.rotor_geometry import RotorGeometry, NREL5MW, DTU10MW, NINPUT
		from rotorse import RPM2RS, RS2RPM, TURBULENCE_CLASS, DRIVETRAIN_TYPE

		rotor = Problem()
		myref = DTU10MW()

		npts_coarse_power_curve = 20 # (Int): number of points to evaluate aero analysis at
		npts_spline_power_curve = 200  # (Int): number of points to use in fitting spline to power curve

		rotor.root = RotorAeroPower(myref, npts_coarse_power_curve, npts_spline_power_curve)
		# ---

		# --- Optimizer
		rotor.driver = ScipyOptimizer()
		rotor.driver.options['optimizer'] = 'SLSQP'
		rotor.driver.options['tol'] = 1.0e-8
		# ---
		# --- Objective
		# AEP0 = 47147596.2911617
		AEP0 = 48113504.25433461
		rotor.driver.add_objective('AEP', scaler=-1./AEP0)
		# --- Design Variables
		rotor.driver.add_desvar('r_max_chord', lower=0.15, upper=0.4) # 0.2366
		rotor.driver.add_desvar('chord_in', low=0.4, high=7.)
		rotor.driver.add_desvar('theta_in', low=-10.0, high=30.0)
		rotor.driver.add_desvar('control_tsr', low=3.0, high=14.0)

		T0 = 1448767.9705024462
		rotor.driver.add_constraint('rated_T',upper=T0*1.02)
		# ---
		# --- Recorder
		rec = DumpRecorder()
		rotor.driver.add_recorder(rec)
		rec.options['record_metadata'] = False
		rec.options['record_unknowns'] = False
		rec.options['record_params'] = False
		rec.options['includes'] = ['r_max_chord', 'chord_in', 'theta_in', 'control_tsr']
		# ---
		# --- Setup
		rotor.setup()

		# === blade grid ===
		rotor['hubFraction'] = myref.hubFraction #0.025  # (Float): hub location as fraction of radius
		rotor['bladeLength'] = myref.bladeLength #61.5  # (Float, m): blade length (if not precurved or swept) otherwise length of blade before curvature
		rotor['precone'] = myref.precone #2.5  # (Float, deg): precone angle
		rotor['tilt'] = myref.tilt #5.0  # (Float, deg): shaft tilt
		# ...
		# ---
		rotor['yaw'] = 0.0  # (Float, deg): yaw error
		rotor['nBlades'] = myref.nBlades #3  # (Int): number of blades

		# === blade geometry ===
		rotor['r_max_chord'] = myref.r_max_chord #0.23577  # (Float): location of max chord on unit radius
		rotor['chord_in'] = myref.chord #np.array([3.2612, 4.5709, 3.3178, 1.4621])  # (Array, m): chord at control points. defined at hub, then at linearly spaced locations from r_max_chord to tip
		rotor['theta_in'] = myref.theta #np.array([13.2783, 7.46036, 2.89317, -0.0878099])  # (Array, deg): twist at control points.  defined at linearly spaced locations from r[idx_cylinder] to tip
		rotor['precurve_in'] = myref.precurve #np.array([0.0, 0.0, 0.0])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
		rotor['presweep_in'] = myref.presweep #np.array([0.0, 0.0, 0.0])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
		rotor['sparT_in'] = myref.spar_thickness #np.array([0.05, 0.047754, 0.045376, 0.031085, 0.0061398])  # (Array, m): spar cap thickness parameters
		rotor['teT_in'] = myref.te_thickness #np.array([0.1, 0.09569, 0.06569, 0.02569, 0.00569])  # (Array, m): trailing-edge thickness parameters

		# === atmosphere ===
		rotor['analysis.rho'] = 1.225  # (Float, kg/m**3): density of air
		rotor['analysis.mu'] = 1.81206e-5  # (Float, kg/m/s): dynamic viscosity of air
		rotor['hub_height'] = myref.hub_height #90.0
		rotor['analysis.shearExp'] = 0.25  # (Float): shear exponent
		rotor['turbine_class'] = myref.turbine_class #TURBINE_CLASS['I']  # (Enum): IEC turbine class
		rotor['cdf_reference_height_wind_speed'] = myref.hub_height #90.0  # (Float): reference hub height for IEC wind speed (used in CDF calculation)

		# === control ===
		rotor['control_Vin'] = myref.control_Vin #3.0  # (Float, m/s): cut-in wind speed
		rotor['control_Vout'] = myref.control_Vout #25.0  # (Float, m/s): cut-out wind speed
		rotor['control_ratedPower'] = myref.rating #5e6  # (Float, W): rated power
		rotor['control_minOmega'] = myref.control_minOmega #0.0  # (Float, rpm): minimum allowed rotor rotation speed
		rotor['control_maxOmega'] = myref.control_maxOmega #12.0  # (Float, rpm): maximum allowed rotor rotation speed
		rotor['control_tsr'] = myref.control_tsr #7.55  # (Float): tip-speed ratio in Region 2 (should be optimized externally)
		rotor['control_pitch'] = myref.control_pitch #0.0  # (Float, deg): pitch angle in region 2 (and region 3 for fixed pitch machines)

		# === aero and structural analysis options ===
		rotor['nSector'] = 4  # (Int): number of sectors to divide rotor face into in computing thrust and power
		rotor['AEP_loss_factor'] = 1.0  # (Float): availability and other losses (soiling, array, etc.)
		rotor['drivetrainType'] = myref.drivetrain #DRIVETRAIN_TYPE['GEARED']  # (Enum)
		# ---

		# --- run and outputs
		rotor.run()
		# ---

		print 'Max Chord Radius = ', rotor['r_max_chord']
		print 'Chord Control Points = ', rotor['chord_in']
		print 'Twist Control Points = ', rotor['theta_in']
		print 'TSR = ', rotor['control_tsr']

		print'----------------'
		print 'Objective = ', -1*rotor['AEP']/AEP0
		print 'AEP = ', rotor['AEP']
		print 'Rated Thrust =', rotor['rated_T']
		print 'Percent change in thrust =', (rotor['rated_T']-T0)/T0 *100





		# import matplotlib.pyplot as plt
		# # plt.plot(rotor['V'], rotor['P']/1e6)
		# # plt.xlabel('Wind Speed (m/s)')
		# # plt.ylabel('Power (MW)')
		# # plt.show()
		# # # ---

		# # -----------------------------------------------------------------------
		# rotor2 = Problem()
		# myref = DTU10MW()

		# npts_coarse_power_curve = 20 # (Int): number of points to evaluate aero analysis at
		# npts_spline_power_curve = 200  # (Int): number of points to use in fitting spline to power curve

		# rotor2.root = RotorAeroPower(myref, npts_coarse_power_curve, npts_spline_power_curve)
		# # ---

		# rotor2.setup()

		# # === blade grid ===
		# rotor2['hubFraction'] = myref.hubFraction #0.025  # (Float): hub location as fraction of radius
		# rotor2['bladeLength'] = myref.bladeLength #61.5  # (Float, m): blade length (if not precurved or swept) otherwise length of blade before curvature
		# rotor2['precone'] = myref.precone #2.5  # (Float, deg): precone angle
		# rotor2['tilt'] = myref.tilt #5.0  # (Float, deg): shaft tilt
		# # ...
		# # ---
		# rotor2['yaw'] = 0.0  # (Float, deg): yaw error
		# rotor2['nBlades'] = myref.nBlades #3  # (Int): number of blades

		# # === blade geometry ===
		# rotor2['r_max_chord'] = myref.r_max_chord #0.23577  # (Float): location of max chord on unit radius
		# rotor2['chord_in'] = myref.chord #np.array([3.2612, 4.5709, 3.3178, 1.4621])  # (Array, m): chord at control points. defined at hub, then at linearly spaced locations from r_max_chord to tip
		# rotor2['theta_in'] = myref.theta #np.array([13.2783, 7.46036, 2.89317, -0.0878099])  # (Array, deg): twist at control points.  defined at linearly spaced locations from r[idx_cylinder] to tip
		# rotor2['precurve_in'] = myref.precurve #np.array([0.0, 0.0, 0.0])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
		# rotor2['presweep_in'] = myref.presweep #np.array([0.0, 0.0, 0.0])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
		# rotor2['sparT_in'] = myref.spar_thickness #np.array([0.05, 0.047754, 0.045376, 0.031085, 0.0061398])  # (Array, m): spar cap thickness parameters
		# rotor2['teT_in'] = myref.te_thickness #np.array([0.1, 0.09569, 0.06569, 0.02569, 0.00569])  # (Array, m): trailing-edge thickness parameters

		# # === atmosphere ===
		# rotor2['analysis.rho'] = 1.225  # (Float, kg/m**3): density of air
		# rotor2['analysis.mu'] = 1.81206e-5  # (Float, kg/m/s): dynamic viscosity of air
		# rotor2['hub_height'] = myref.hub_height #90.0
		# rotor2['analysis.shearExp'] = 0.25  # (Float): shear exponent
		# rotor2['turbine_class'] = myref.turbine_class #TURBINE_CLASS['I']  # (Enum): IEC turbine class
		# rotor2['cdf_reference_height_wind_speed'] = myref.hub_height #90.0  # (Float): reference hub height for IEC wind speed (used in CDF calculation)

		# # === control ===
		# rotor2['control_Vin'] = myref.control_Vin #3.0  # (Float, m/s): cut-in wind speed
		# rotor2['control_Vout'] = myref.control_Vout #25.0  # (Float, m/s): cut-out wind speed
		# rotor2['control_ratedPower'] = myref.rating #5e6  # (Float, W): rated power
		# rotor2['control_minOmega'] = myref.control_minOmega #0.0  # (Float, rpm): minimum allowed rotor rotation speed
		# rotor2['control_maxOmega'] = myref.control_maxOmega #12.0  # (Float, rpm): maximum allowed rotor rotation speed
		# rotor2['control_tsr'] = myref.control_tsr #7.55  # (Float): tip-speed ratio in Region 2 (should be optimized externally)
		# rotor2['control_pitch'] = myref.control_pitch #0.0  # (Float, deg): pitch angle in region 2 (and region 3 for fixed pitch machines)

		# # === aero and structural analysis options ===
		# rotor2['nSector'] = 4  # (Int): number of sectors to divide rotor face into in computing thrust and power
		# rotor2['AEP_loss_factor'] = 1.0  # (Float): availability and other losses (soiling, array, etc.)
		# rotor2['drivetrainType'] = myref.drivetrain #DRIVETRAIN_TYPE['GEARED']  # (Enum)
		# # ---

		# # --- run and outputs
		# rotor2.run()
		# # -----------------------------------------------------------------------


		outpath = '..\..\..\docs\images'
		# Power Curve
		f, ax = plt.subplots(1,1,figsize=(5.3, 4))
		ax.plot(rotor['V'], rotor['P']/1e6)
		ax.set(xlabel='Wind Speed (m/s)' , ylabel='Power (MW)')
		ax.set_ylim([0, 10.3])
		ax.set_xlim([0, 25])
		f.tight_layout()
		ax.spines['right'].set_visible(False)
		ax.spines['top'].set_visible(False)
		# f.savefig(os.path.abspath(os.path.join(outpath,'power_curve_dtu10mw.png')))
		# f.savefig(os.path.abspath(os.path.join(outpath,'power_curve_dtu10mw.pdf')))

		# Chord
		fc, axc = plt.subplots(1,1,figsize=(5.3, 4))
		rc_c = np.r_[0.0, myref.r_cylinder, np.linspace(rotor['r_max_chord'], 1.0, NINPUT-2)]
		rc_c2 = np.r_[0.0, myref.r_cylinder, np.linspace(rotor2['r_max_chord'], 1.0, NINPUT-2)]
		r = (rotor['r_pts'] - rotor['Rhub'])/rotor['bladeLength']
		# axc.plot(r, rotor2['chord'], c='k', label='Initial')
		# axc.plot(rc_c2, rotor2['chord_in'], '.', c='k')
		axc.plot(r, rotor['chord'], c='b', label='Optimized')
		axc.plot(rc_c, rotor['chord_in'], '.', c='b')
		# for i, (x, y) in enumerate(zip(rc_c, rotor['chord_in'])):
		#     txt = '$c_%d$' % i
		#     if i<=1:
		#         axc.annotate(txt, (x,y), xytext=(x+0.01,y-0.4), textcoords='data')
		#     else:
		#         axc.annotate(txt, (x,y), xytext=(x+0.01,y+0.2), textcoords='data')
		# for i, (x, y) in enumerate(zip(rc_c2, rotor2['chord_in'])):
		#     txt = '$c_%d$' % i
		#     if i==0:
		#         axc.annotate(txt, (x,y), xytext=(x+0.01,y-0.4), textcoords='data', color='blue')
		#     else:
		#         axc.annotate(txt, (x,y), xytext=(x+0.01,y+0.2), textcoords='data', color='blue')
		axc.set(xlabel='Blade Fraction, $r/R$' , ylabel='Chord (m)')
		axc.set_ylim([0, 7.5])
		axc.set_xlim([0, 1.1])
		fc.tight_layout()
		axc.spines['right'].set_visible(False)
		axc.spines['top'].set_visible(False)
		# fc.savefig(os.path.abspath(os.path.join(outpath,'chord_opt_10mw.png')))
		# fc.savefig(os.path.abspath(os.path.join(outpath,'chord_opt_10mw.pdf')))

		# Twist
		ft, axt = plt.subplots(1,1,figsize=(5.3, 4))
		rc_t = rc_c#np.linspace(myref.r_cylinder, 1.0, NINPUT)
		rc_t2 = rc_c2#np.linspace(myref.r_cylinder, 1.0, NINPUT)
		r = (rotor['r_pts'] - rotor['Rhub'])/rotor['bladeLength']
		# axt.plot(r, rotor2['theta'], c='k', label='Initial')
		# axt.plot(rc_t2, rotor2['theta_in'], '.', c='k')
		axt.plot(r, rotor['theta'], c='b', label='Optimized')
		axt.plot(rc_t, rotor['theta_in'], '.', c='b')
		# for i, (x, y) in enumerate(zip(rc_t, rotor['theta_in'])):
		#     txt = '$\Theta_%d$' % i
		#     axt.annotate(txt, (x,y), xytext=(x+0.01,y+0.1), textcoords='data')
		# for i, (x, y) in enumerate(zip(rc_t2, rotor2['theta_in'])):
		#     txt = '$\Theta_%d$' % i
		#     axt.annotate(txt, (x,y), xytext=(x+0.01,y+0.1), textcoords='data', color='blue')
		axt.set(xlabel='Blade Fraction, $r/R$' , ylabel='Twist ($\deg$)')
		axt.set_ylim([-4, 16])
		axt.set_xlim([0, 1.1])
		ft.tight_layout()
		axt.spines['right'].set_visible(False)
		axt.spines['top'].set_visible(False)
		# ft.savefig(os.path.abspath(os.path.join(outpath,'theta_opt_10mw.png')))
		# ft.savefig(os.path.abspath(os.path.join(outpath,'theta_opt_10mw.pdf')))

		plt.show()


if __name__ == "__main__":

	rotor = RotorSE_Example2()
	rotor.execute()


import copy
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

import numpy as np
from scipy.interpolate import PchipInterpolator, Akima1DInterpolator, interp1d
from wisdem.rotorse.rotor_geometry_yaml import remap2grid, remapAirfoil, arc_length, rotate, ReferenceBlade

def plot_spanwise_vars(blade, path, show_plots = True):
    
    # Chord
    fc, axc  = plt.subplots(1,1,figsize=(2.5, 2))
    fc.subplots_adjust(bottom=0.15, left=0.25)
    axc.plot(blade['pf']['s'], blade['pf']['chord'])
    axc.set(xlabel='r/R' , ylabel='Chord (m)')
    axc.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_chord.png'
    fc.savefig(path + fig_name)
    
    # Theta
    ft, axt  = plt.subplots(1,1,figsize=(2.5, 2))
    ft.subplots_adjust(bottom=0.15, left=0.25)
    axt.plot(blade['pf']['s'], blade['pf']['theta'])
    axt.set(xlabel='r/R' , ylabel='Twist (deg)')
    axt.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_theta.png'
    ft.savefig(path + fig_name)
    
    # Pitch axis
    fp, axp  = plt.subplots(1,1,figsize=(2.5, 2))
    fp.subplots_adjust(bottom=0.15, left=0.25)
    axp.plot(blade['pf']['s'], blade['pf']['p_le']*100.)
    axp.set(xlabel='r/R' , ylabel='Pitch Axis (%)')
    axp.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_p_le.png'
    fp.savefig(path + fig_name)
    
    
    # Planform
    le = blade['pf']['p_le']*blade['pf']['chord']
    te = (1. - blade['pf']['p_le'])*blade['pf']['chord']

    fpl, axpl  = plt.subplots(1,1,figsize=(2.5, 2))
    fpl.subplots_adjust(bottom=0.15, left=0.25)
    axpl.plot(blade['pf']['s'], -le)
    axpl.plot(blade['pf']['s'], te)
    axpl.set(xlabel='r/R' , ylabel='Planform (m)')
    axpl.grid(color=[0.8,0.8,0.8], linestyle='--')
    axpl.legend()
    fig_name = 'design_planform_planform.png'
    fpl.savefig(path + fig_name)
    
    # Relative thickness
    frt, axrt  = plt.subplots(1,1,figsize=(2.5, 2))
    frt.subplots_adjust(bottom=0.15, left=0.25)
    axrt.plot(blade['pf']['s'], blade['pf']['rthick']*100.)
    axrt.set(xlabel='r/R' , ylabel='Relative Thickness (%)')
    axrt.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_rthick.png'
    frt.savefig(path + fig_name)
    
    # Absolute thickness
    fat, axat  = plt.subplots(1,1,figsize=(2.5, 2))
    fat.subplots_adjust(bottom=0.15, left=0.25)
    axat.plot(blade['pf']['s'], blade['pf']['rthick']*blade['pf']['chord'])
    axat.set(xlabel='r/R' , ylabel='Absolute Thickness (m)')
    axat.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_absthick.png'
    fat.savefig(path + fig_name)
    
    # Prebend
    fpb, axpb  = plt.subplots(1,1,figsize=(2.5, 2))
    fpb.subplots_adjust(bottom=0.15, left=0.25)
    axpb.plot(blade['pf']['s'], blade['pf']['precurve'])
    axpb.set(xlabel='r/R' , ylabel='Prebend (m)')
    axpb.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_prebend.png'
    fpb.savefig(path + fig_name)
    
    # Sweep
    fsw, axsw  = plt.subplots(1,1,figsize=(2.5, 2))
    fsw.subplots_adjust(bottom=0.15, left=0.25)
    axsw.plot(blade['pf']['s'], blade['pf']['presweep'])
    axsw.set(xlabel='r/R' , ylabel='Presweep (m)')
    axsw.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_presweep.png'
    plt.subplots_adjust(left = 0.14)
    fsw.savefig(path + fig_name)
    
    idx_spar  = [i for i, sec in enumerate(blade['st']['layers']) if sec['name'].lower()==self.spar_var[0].lower()][0]
    idx_te    = [i for i, sec in enumerate(blade['st']['layers']) if sec['name'].lower()==self.te_var.lower()][0]
    idx_skin  = [i for i, sec in enumerate(blade['st']['layers']) if sec['name'].lower()=='shell_skin'][0]
    
    # Spar caps thickness
    fsc, axsc  = plt.subplots(1,1,figsize=(2.5, 2))
    fsc.subplots_adjust(bottom=0.15, left=0.25)
    axsc.plot(blade['st']['layers'][idx_spar]['thickness']['grid'], blade['st']['layers'][idx_spar]['thickness']['values'])
    axsc.set(xlabel='r/R' , ylabel='Spar Caps Thickness (m)')
    axsc.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_sc.png'
    plt.subplots_adjust(left = 0.14)
    fsc.savefig(path + fig_name)
    
    # TE reinf thickness
    fte, axte  = plt.subplots(1,1,figsize=(2.5, 2))
    fte.subplots_adjust(bottom=0.15, left=0.25)
    axte.plot(blade['st']['layers'][idx_te]['thickness']['grid'], blade['st']['layers'][idx_te]['thickness']['values'])
    axte.set(xlabel='r/R' , ylabel='TE Reinf. Thickness (m)')
    axte.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_te.png'
    plt.subplots_adjust(left = 0.14)
    fte.savefig(path + fig_name)
    
    # Skin
    fsk, axsk  = plt.subplots(1,1,figsize=(2.5, 2))
    fsk.subplots_adjust(bottom=0.15, left=0.25)
    axsk.plot(blade['st']['layers'][idx_skin]['thickness']['grid'], blade['st']['layers'][idx_skin]['thickness']['values'])
    axsk.set(xlabel='r/R' , ylabel='Shell Skin Thickness (m)')
    axsk.grid(color=[0.8,0.8,0.8], linestyle='--')
    fig_name = 'design_planform_skin.png'
    fsk.savefig(path + fig_name)
    
    if show_plots:
        plt.show()
    return None

###################################

def find_half(arc_pos, arc, LE_idx, Naf):
    if arc_pos >= arc[LE_idx]:
        idx_half = range(LE_idx, Naf)
    else:
        idx_half = range(0, LE_idx)
    return idx_half

def arc2xy_section(arc, x, y, start_nd_arc, end_nd_arc, LE_idx, Naf, Nsec):

    # find SS or PS to interpolate over
    idx_half_start = find_half(start_nd_arc, arc, LE_idx, Naf)
    idx_half_end   = find_half(end_nd_arc, arc, LE_idx, Naf)

    # No LE or Te crossing of composite section
    # if idx_half_start == idx_half_end:
    section_x = remap2grid(arc[idx_half_start], x[idx_half_start], np.linspace(start_nd_arc, end_nd_arc, num = Nsec))#, spline=interp1d)
    section_y = remap2grid(x[idx_half_start], y[idx_half_start], section_x)#, spline=interp1d)

    # else:
        # print('LE / TE not yet implimented')

    return section_x, section_y

def arc2xy_web(arc, x, y, start_nd_arc, end_nd_arc, LE_idx, Naf, Nsec):

    # find SS or PS to interpolate over
    idx_half_start = find_half(start_nd_arc, arc, LE_idx, Naf)
    idx_half_end   = find_half(end_nd_arc, arc, LE_idx, Naf)

    # No LE or Te crossing of composite section
    x_start = remap2grid(arc[idx_half_start], x[idx_half_start], np.linspace(start_nd_arc, end_nd_arc, num = Nsec), spline=interp1d)
    y_start = remap2grid(x[idx_half_start], y[idx_half_start], x_start, spline=interp1d)
    x_end   = remap2grid(arc[idx_half_end], x[idx_half_end], np.linspace(start_nd_arc, end_nd_arc, num = Nsec), spline=interp1d)
    y_end   = remap2grid(x[idx_half_end], y[idx_half_end], x_end, spline=interp1d)

    return [x_start, x_end], [y_start, y_end]


def plot_lofted(blade, layer_vars, path, show_plots = True):

    NPTS = len(blade['pf']['s'])
    Naf  = len(blade['profile'][:,0,0])
    Nsec = 30
    

    
    
    # Get Profile Coordinates, xy and arc
    out = {}
    out['profile']           = {}
    out['profile']['xy']     = np.flip(copy.deepcopy(blade['profile']), axis=0)
    out['profile']['LE_idx'] = np.array([np.argmin(out['profile']['xy'][:,0,i]) for i in range(NPTS)])
    out['profile']['arc']    = []
    for i in range(NPTS):
        profile_i = out['profile']['xy'][:,:,i]
        # if profile_i[0,1] != profile_i[-1,1]:
        #     profile_i = np.row_stack((profile_i, profile_i[0,:])) 
        arc = arc_length(profile_i[:,0], profile_i[:,1])
        out['profile']['arc'].append(arc/arc[-1])

    # For each layer, get surface locations
    for var in layer_vars:
        type_sec = 'layers'
        idx_sec  = [i for i, sec in enumerate(blade['st'][type_sec]) if sec['name'].lower()==var.lower()][0]
        out[var] = {}
        out[var]['start_nd_arc'] = np.array(blade['st'][type_sec][idx_sec]['start_nd_arc']['values'])
        out[var]['end_nd_arc']   = np.array(blade['st'][type_sec][idx_sec]['end_nd_arc']['values'])

        out[var]['xy'] = np.zeros((Nsec,2,NPTS))
        for i in range(NPTS):
            profile_i = out['profile']['xy'][:,:,i]
            # if profile_i[0,1] != profile_i[-1,1]:
            #     profile_i = np.row_stack((profile_i, profile_i[0,:]))
            out[var]['xy'][:,:,i] = np.column_stack(arc2xy_section(out['profile']['arc'][i], profile_i[:,0], profile_i[:,1], out[var]['start_nd_arc'][i], out[var]['end_nd_arc'][i], out['profile']['LE_idx'][i], Naf, Nsec))

    # For each web, get locations

    # Other locations (LE, pitch axis, etc)

    # Loop through plot elements, span; translate
    plt_vars = ['profile']
    plt_vars.extend(layer_vars)
    
    
    for i in range(NPTS):
        for j, var in enumerate(plt_vars):
            # print(var,i)

            # center about pitch axis
            out[var]['xy'][:,0,i] = out[var]['xy'][:,0,i] - blade['pf']['p_le'][i]

            # scale by chord
            out[var]['xy'][:,:,i] = out[var]['xy'][:,:,i] * blade['pf']['chord'][i]

            # rotate
            out[var]['xy'][:,:,i] = np.column_stack(rotate(0., 0., out[var]['xy'][:,0,i], out[var]['xy'][:,1,i], np.radians(blade['pf']['theta'][i])))

            # prebend translation
            out[var]['xy'][:,1,i] = out[var]['xy'][:,1,i] - blade['pf']['precurve'][i]

            # sweep translation
            out[var]['xy'][:,0,i] = out[var]['xy'][:,0,i] - blade['pf']['presweep'][i]
            

        
        

    # Plotting

    ######### 2D
    color    = ['k','b', 'r']
    for i in range(NPTS):
        for j, var in enumerate(plt_vars):
            if var == 'profile':
                profile_i = out['profile']['xy'][:,:,i]
                if profile_i[0,1] != profile_i[-1,1]:
                    profile_i = np.row_stack((profile_i, profile_i[0,:]))
                plt.plot(profile_i[:,0], profile_i[:,1], color=color[j])
            else:
                # # if blade['st'][type_sec][20]['thickness']['values'][i] != None and blade['st'][type_sec][21]['thickness']['values'][i] != None:
                plt.plot(out[var]['xy'][:,0,i], out[var]['xy'][:,1,i], color=color[j])
                plt.plot([out[var]['xy'][0,0,i],out[var]['xy'][-1,0,i]], [out[var]['xy'][0,1,i],out[var]['xy'][-1,1,i]], '.', color=color[j])
    plt.show()
    plt.axis('equal')
        


    ######### 3D

    color    = ['k','b', 'r']

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    for i in range(NPTS):
        for j, var in enumerate(plt_vars):
            if var == 'profile':
                profile_i = out['profile']['xy'][:,:,i]
                if profile_i[0,1] != profile_i[-1,1]:
                    profile_i = np.row_stack((profile_i, profile_i[0,:]))
                Y = profile_i[:,0]
                Z = profile_i[:,1]
                X = [blade['pf']['r'][i]]*len(profile_i[:,0])
                ax.plot(X, Y, zs=Z, color=color[j])
            else:
                plt.plot(out[var]['xy'][:,0,i], out[var]['xy'][:,1,i], color=color[j])
                plt.plot([out[var]['xy'][0,0,i],out[var]['xy'][-1,0,i]], [out[var]['xy'][0,1,i],out[var]['xy'][-1,1,i]], '.', color=color[j])



    Y = out['profile']['xy'][:,0,0]
    Z = out['profile']['xy'][:,1,0]
    X = np.array([blade['pf']['r']]*len(Y))
    max_range = np.array([X.max()-X.min(), Y.max()-Y.min(), Z.max()-Z.min()]).max() / 2.0
    mid_x = (X.max()+X.min()) * 0.5
    mid_y = (Y.max()+Y.min()) * 0.5
    mid_z = (Z.max()+Z.min()) * 0.5
    ax.set_xlim(mid_x - max_range, mid_x + max_range)
    ax.set_ylim(mid_y - max_range, mid_y + max_range)
    ax.set_zlim(mid_z - max_range, mid_z + max_range)


    ax.view_init(azim=-90., elev=-180.)
    plt.show()

    return None


if __name__ == "__main__":

    fname_input       = "/mnt/c/Users/egaertne/WISDEM/nrel15mw/design/turbine_inputs/NREL15MW_opt_v03.yaml"
    dir_out           = "/mnt/c/Users/egaertne/WISDEM/nrel15mw/design/outputs/NREL15MW_opt_v03/post"

    ## Load and Format Blade
    refBlade = ReferenceBlade()
    refBlade.verbose  = True
    refBlade.spar_var = ['Spar_cap_ss', 'Spar_cap_ps']
    refBlade.te_var   = 'TE_reinforcement'
    refBlade.NINPUT   = 8
    refBlade.NPITS    = 50
    # refBlade.NPTS_AfProfile = 200
    refBlade.validate = False
    # refBlade.fname_schema = "turbine_inputs/IEAontology_schema.yaml"

    blade = refBlade.initialize(fname_input)

    plt_vars = ['Spar_cap_ss', 'Spar_cap_ps']
    plot_lofted(blade, plt_vars, dir_out)



from openmdao.api import Group, ExplicitComponent
from .csystem import DirectionVector
from .utilities import interp_with_deriv
from wisdem.commonse import NFREQ
import numpy as np


# class TowerModes(ExplicitComponent):
#     def setup(self):

#         self.add_input('tower_freq', val=np.zeros(2), units='Hz', desc='First natural frequencies of tower (and substructure)')
#         self.add_input('rotor_omega', val=0.0, units='rpm', desc='rated rotor rotation speed')
#         self.add_input('gamma_freq', val=0.0, desc='partial safety factor for fatigue')
#         self.add_discrete_input('blade_number', 3, desc='number of rotor blades')

#         self.add_output('frequencyNP_margin_low', val=np.zeros(2), desc='Upper bound constraint of tower/structure frequency to blade passing frequency with margin')
#         self.add_output('frequencyNP_margin_high', val=np.zeros(2), desc='Lower bound constraint of tower/structure frequency to blade passing frequency with margin')
#         self.add_output('frequency1P_margin_low', val=np.zeros(2), desc='Upper bound constraint of tower/structure frequency to rotor frequency with margin')
#         self.add_output('frequency1P_margin_high', val=np.zeros(2), desc='Lower bound constraint of tower/structure frequency to rotor frequency with margin')

#         # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
#     def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
#         freq_struct = inputs['tower_freq']
#         gamma       = inputs['gamma_freq']
#         oneP        = (inputs['rotor_omega']/60.0)
#         oneP_high   = oneP * gamma
#         oneP_low    = oneP / gamma
#         threeP      = oneP * discrete_inputs['blade_number']
#         threeP_high = threeP * gamma
#         threeP_low  = threeP / gamma
        
#         # Compute margins between (N/3)P and structural frequencies
#         indicator_high = threeP_high * np.ones(freq_struct.shape)
#         indicator_high[freq_struct < threeP_low] = 1e-16
#         outputs['frequencyNP_margin_high'] = freq_struct / indicator_high

#         indicator_low = threeP_low * np.ones(freq_struct.shape)
#         indicator_low[freq_struct > threeP_high] = 1e30
#         outputs['frequencyNP_margin_low']  = freq_struct / indicator_low

#         # Compute margins between 1P and structural frequencies
#         indicator_high = oneP_high * np.ones(freq_struct.shape)
#         indicator_high[freq_struct < oneP_low] = 1e-16
#         outputs['frequency1P_margin_high'] = freq_struct / indicator_high

#         indicator_low = oneP_low * np.ones(freq_struct.shape)
#         indicator_low[freq_struct > oneP_high] = 1e30
#         outputs['frequency1P_margin_low']  = freq_struct / indicator_low

class TowerModes(ExplicitComponent):
    def setup(self):

        self.add_input('tower_freq', val=np.zeros(2), units='Hz', desc='First natural frequencies of tower (and substructure)')
        self.add_input('rotor_omega', val=0.0, units='rpm', desc='rated rotor rotation speed')
        self.add_input('gamma_freq', val=0.0, desc='partial safety factor for fatigue')
        self.add_discrete_input('blade_number', 3, desc='number of rotor blades')

        self.add_output('frequencyNP_margin', val=np.zeros(2), desc='constraint on tower/structure frequency to blade passing frequency with margin')
        self.add_output('frequency1P_margin', val=np.zeros(2), desc='constraint on tower/structure frequency to rotor frequency with margin')

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        freq_struct = inputs['tower_freq']
        gamma       = inputs['gamma_freq']
        oneP        = (inputs['rotor_omega']/60.0)
        oneP_high   = oneP * gamma
        oneP_low    = oneP / gamma
        threeP      = oneP * discrete_inputs['blade_number']
        threeP_high = threeP * gamma
        threeP_low  = threeP / gamma
        
        outputs['frequencyNP_margin'] = np.array([min([f-threeP_low, threeP_high-f]) for f in freq_struct]).flatten()
        outputs['frequency1P_margin'] = np.array([min([f-oneP_low, oneP_high-f]) for f in freq_struct]).flatten()


class MaxTipDeflection(ExplicitComponent):
    def initialize(self):
        self.options.declare('nFullTow')
        
    def setup(self):
        nFullTow = self.options['nFullTow']

        self.add_discrete_input('downwind',       val=False)
        self.add_input('tip_deflection', val=0.0,       units='m',  desc='Blade tip deflection in yaw x-direction')
        self.add_input('Rtip',           val=0.0,       units='m',  desc='Blade tip location in z_b')
        self.add_input('precurveTip',    val=0.0,       units='m',  desc='Blade tip location in x_b')
        self.add_input('presweepTip',    val=0.0,       units='m',  desc='Blade tip location in y_b')
        self.add_input('precone',        val=0.0,       units='deg',desc='Rotor precone angle')
        self.add_input('tilt',           val=0.0,       units='deg',desc='Nacelle uptilt angle')
        self.add_input('overhang',       val=0.0,       units='m',  desc='Horizontal distance between hub and tower-top axis')
        self.add_input('z_full',         val=np.zeros(nFullTow),units='m',  desc='z-coordinates of tower at fine-section nodes')
        self.add_input('d_full',         val=np.zeros(nFullTow),units='m',  desc='Diameter of tower at fine-section nodes')
        self.add_input('gamma_m',        val=0.0, desc='safety factor on materials')

        self.add_output('tip_deflection_ratio',      val=0.0,           desc='Ratio of blade tip deflectiion towardsa the tower and clearance between undeflected blade tip and tower')
        self.add_output('blade_tip_tower_clearance', val=0.0, units='m',desc='Clearance between undeflected blade tip and tower in x-direction of yaw c.s.')

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack variables
        z_tower     = inputs['z_full']
        d_tower     = inputs['d_full']
        overhang    = inputs['overhang']
        tt2hub      = overhang * np.sin(inputs['tilt'] / 180. * np.pi)
        precone     = inputs['precone']
        tilt        = inputs['tilt']
        delta       = inputs['tip_deflection']
        upwind      = not discrete_inputs['downwind']
        
        # Coordinates of blade tip in yaw c.s.
        blade_yaw = DirectionVector(inputs['precurveTip'], inputs['presweepTip'], inputs['Rtip']).\
                    bladeToAzimuth(precone).azimuthToHub(180.0).hubToYaw(tilt)

        # Find the radius of tower where blade passes
        z_interp = z_tower[-1] + tt2hub + blade_yaw.z
        
        d_interp, ddinterp_dzinterp, ddinterp_dtowerz, ddinterp_dtowerd = interp_with_deriv(z_interp, z_tower, d_tower)
        r_interp = 0.5 * d_interp
        drinterp_dzinterp = 0.5 * ddinterp_dzinterp
        drinterp_dtowerz  = 0.5 * ddinterp_dtowerz
        drinterp_dtowerd  = 0.5 * ddinterp_dtowerd

        # Max deflection before strike
        if upwind:
            parked_margin = overhang - blade_yaw.x - r_interp
        else:
            parked_margin = -overhang + blade_yaw.x - r_interp
        outputs['blade_tip_tower_clearance']   = parked_margin
        outputs['tip_deflection_ratio']        = delta * inputs['gamma_m'] / parked_margin

    
class TurbineConstraints(Group):

    def initialize(self):
        self.options.declare('nFull')
        
    def setup(self):
        nFull = self.options['nFull']
        
        self.add_subsystem('modes', TowerModes(), promotes=['*'])
        self.add_subsystem('tipd', MaxTipDeflection(nFullTow=nFull), promotes=['*'])



# Utility functions used by drivese components

import numpy as np
import scipy as scp
from math import pi, cos, sqrt, sin, exp, log10, log
import sys

#-------------------------------------------------------------------------
# Supporting functions

#If user does not specify key information about turbine properties,
#they are estimated from curve fits to basic turbine configuration parameters

# if user inputs forces and zero rotor mass then rotor mass estimated
def get_rotor_mass(machine_rating, deriv):  
    out = [23.566 * machine_rating]
    if deriv:
        out.extend([23.566])
    return out

# moments taken to scale approximately with force (rotor mass) and distance (distance_hub2mb)
def get_My(rotor_mass, distance_hub2mb):
    if distance_hub2mb == 0:
        # approximate rotor diameter from rotor mass
        distance_hub2mb = get_distance_hub2mb((rotor_mass + 49089) / 1170.6)
    return 59.7 * rotor_mass * distance_hub2mb

# moments taken to scale roughly with force (rotor mass) and distance (distance_hub2mb)
def get_Mz(rotor_mass, distance_hub2mb):  
    if distance_hub2mb == 0:
        # approximate rotor diameter from rotor mass
        distance_hub2mb = get_distance_hub2mb((rotor_mass - 49089) / 1170.6)
    return 53.846 * rotor_mass * distance_hub2mb

# function to estimate the location of the main bearing location from the hub center
def get_distance_hub2mb(rotor_diameter, deriv=False):
    out = 0.007835 * rotor_diameter + 0.9642
    # Ignoring deriv calculation for now
    #if deriv:
    #    out.extend([.007835])
    return out

# -------------------------------------------------
# Bearing support functions
# returns facewidth, mass for bearings without fatigue analysis
def resize_for_bearings(D_shaft, type, deriv):
    # assume low load rating for bearing
    if type == 'CARB':  # p = Fr, so X=1, Y=0
        out = [D_shaft, .2663 * D_shaft + .0435, 1561.4 * D_shaft**2.6007]
        if deriv == True:
            out.extend([1., .2663, 1561.4 * 2.6007 * D_shaft**1.6007])
    elif type == 'SRB':
        out = [D_shaft, .2762 * D_shaft, 876.7 * D_shaft**1.7195]
        if deriv == True:
            out.extend([1., .2762, 876.7 * 1.7195 * D_shaft**0.7195])
    elif type == 'TRB1':
        out = [D_shaft, .0740, 92.863 * D_shaft**.8399]
        if deriv == True:
            out.extend([1., 0., 92.863 * 0.8399 * D_shaft**(0.8399 - 1.)])
    elif type == 'CRB':
        out = [D_shaft, .1136 * D_shaft, 304.19 * D_shaft**1.8885]
        if deriv == True:
            out.extend([1., .1136, 304.19 * 1.8885 * D_shaft**0.8885])
    elif type == 'TRB2':
        out = [D_shaft, .1499 * D_shaft, 543.01 * D_shaft**1.9043]
        if deriv == True:
            out.extend([1., .1499, 543.01 * 1.9043 * D_shaft**.9043])
    elif type == 'RB':  # factors depend on ratio Fa/C0, C0 depends on bearing... TODO: add this functionality
        out = [D_shaft, .0839, 229.47 * D_shaft**1.8036]
        if deriv == True:
            out.extend([1.0, 0.0, 229.47 * 1.8036 * D_shaft**0.8036])

    # shaft diameter, facewidth, mass. if deriv==True, provides derivatives.
    return out


#%%---------------------------------------------------------------
# Revised mainshaft flange calculations - added July 2019
    
def mainshaftFlangeCalc(D_in, D_face_MB, flange_thickness, debug=False):
    ''' Compute mass, center of mass, and cost of mainshaft flange
    
        Rxx identifiers in comments are the row numbers in the original Excel spreadsheet
    
        Parameters
        ----------
        D_in : float
          Mainshaft Inner Bore Diameter (from DriveSE == D_max * shaft_ratio (0.10))
        D_face_MB : float
          Mainshaft Up-Wind Bearing Face Diameter (from DriveSE == D_max ??)
        flange_thickness : float
          Mainshaft Flange Thickness (from HubSE - SphHub::main_flange_thick ??)
          
        Returns
        -------
        flange_length : float
          flange length in m
        mass_flange : float
          flange mass in kg
        cm_flange : float
          flange center of mass location in m
        cost_flange : float
          flange cost in $
    '''
	
    PCT_SHOULDER_HT_ID = 3.15     # R12 Main Bearing Shaft Shoulder Height (as a percentage of ID)  3.150  %
    RATIO_SHOULDER_W2H = 6        # R15 Main Bearing Shaft Shoulder Width to Height Ratio  6.000  
    RATIO_SHOULDER_STEPDOWN = 3   # R17 Main Bearing Shaft Shoulder Width Step down ratio  3.000  
    RATIO_R2D = 0.3               # R20 r/d ratio (Petersons SCF Guide)  0.30  
    FLANGE_BOLT_SIZE = 0.048      # R23 Flange Bolt Size (Diameter)  0.048  m
    FLANGE_BOLT_DIAM_INCR = 0.003 # R24 Flange Bolt Diameter Increase  0.003  m
    RATIO_DIST2BHD = 1.5          # R26 Distance from Bolt Circle Center to Flange Edge (as ratio of Bolt Hole Diameter)  1.500  
    DENSITY_42CrMo4 = 7800        # R36 Density of Forging (42CrMo4 Steel)  7800  kg/m3
    DENSITY_S355    = 7850        # R37 Density of Structural Steel (S355 Steel)  7850  kg/m3
    COST_FORGING = 3.5            # R61 Mainshaft Flange Forging Cost (42CrMo4)  3.5  USD/kg
    COST_LOCKPLATE = 3.0          # R62 Rotor Lock Plate Cost (S355)  3  USD/kg
    
    density_forging = DENSITY_42CrMo4
    
    # Mainshaft Flange Geometry                                                                                           
                                                                                      #      
    shoulder_diam = D_face_MB * ((100 + PCT_SHOULDER_HT_ID) / 100)                    #  R13 Main Bearing Shaft Shoulder Height Diameter  1.155  m
    shoulder_height = (shoulder_diam - D_face_MB) / 2                                 #  R14 Main Bearing Shaft Shoulder Height  0.018  m
    shoulder_width = shoulder_height * RATIO_SHOULDER_W2H                             #  R16 Main Bearing Shaft Shoulder Width  0.106  m
    shoulder_stepdown_width = shoulder_width / RATIO_SHOULDER_STEPDOWN                #  R18 Main Bearing Shaft Shoulder Width Step down Width  0.035  m
    tot_shoulder_width = shoulder_width + shoulder_stepdown_width                     #      
                                                                                      #   
    flange_trans_radius = D_face_MB * RATIO_R2D                                       #  R21 Mainshaft to Mainshaft Flange Transition Radius  0.336  m
    flange_length = shoulder_width + shoulder_stepdown_width \
                  + flange_trans_radius + flange_thickness                            #  R30 Mainshaft Flange Length (Total)  0.817  m
    
    bolt_circle_ID = D_face_MB + (2 * flange_trans_radius)                            #  R22 Flange Bolt Circle  ID  1.792  m
    bolt_hole_diam = FLANGE_BOLT_SIZE + FLANGE_BOLT_DIAM_INCR                         #  R25 Flange Bolt Hole Diameter  0.051  m
    bolt_circ_flange_len = 2 * (RATIO_DIST2BHD * bolt_hole_diam)                      #  R27 Bolt Circle Flange Length (Distance between ID and OD)  0.153  m
    bolt_circle_OD = bolt_circle_ID + (2 * bolt_circ_flange_len)                      #  R28 Bolt Circle OD (Mainshaft OD)  2.098  m
                                                                                      #
    # Mainshaft Flange Design Allowable and Material Properties                       #
    # Mainshaft Flange Mass Calculations                                              #
    # Mainshaft Flange Volume(s) - subcomponents   
                                       #
    vol_shoulder  = (pi/4) * tot_shoulder_width  * (shoulder_diam**2 - D_in**2)       #  R40 Mainshaft Volume (Main bearing Shoulder)  0.120  m3
    vol_tzone1    = (pi/4) * flange_trans_radius * (D_face_MB**2     - D_in**2)       #  R41 Mainshaft Volume (Transition zone 1-base)  0.265  m3
    vol_tzone2    = (1 - pi/4) * (pi/4) * flange_trans_radius \
                             * ((D_face_MB + flange_trans_radius)**2 - D_in**2)       #  R42 Mainshaft Volume (Transition zone 2-fillet)  0.106  m3
    vol_bcflange  = (pi/4) * flange_thickness *   (bolt_circle_OD**2 - D_in**2)       #  R43 Mainshaft Volume (Bolt Circle Flange)  1.109  m3
    vol_flange = vol_shoulder + vol_tzone1 + vol_tzone2 + vol_bcflange                #  R44 Mainshaft Volume (TOTAL)  1.600  m3
                                                                                      #   
    mass_flange = vol_flange * density_forging                                        #  R46 Mainshaft Flange Mass (TOTAL)  12,478.4  kg
                                                                                      #  
    # Mainshaft Flange Centroid Calculations 
                                             #  
    mass_shoulder = vol_shoulder * density_forging                                    #  R49 Mainshaft Mass (Main bearing Shoulder)  937.7  kg
    mass_tzone1   = vol_tzone1   * density_forging                                    #  R51 Mainshaft Mass (Transition zone 1-base)  2,067.4  kg
    mass_tzone2   = vol_tzone2   * density_forging                                    #  R53 Mainshaft Mass (Transition zone 2-fillet)  826.0  kg
    mass_bcflange = vol_bcflange * density_forging                                    #  R55 Mainshaft Mass (Bolt Circle Flange)  8,647.3  kg
                                                                    
    cm_shoulder = 0.5 * tot_shoulder_width                                            #  R50 Center of Mass (Main bearing Shoulder)    0.071  m
    cm_tzone1   = tot_shoulder_width + (flange_trans_radius/2)                        #  R52 Center of Mass (Transition zone 1-base)  0.309  m
    cm_tzone2   = tot_shoulder_width + (2 * flange_trans_radius/3)                    #  R54 Center of Mass (Transition zone 2-fillet)   0.365  m
    cm_bcflange = tot_shoulder_width + (flange_thickness/2) + flange_trans_radius     #  R56 Center of Mass (Bolt Circle Flange)  0.647  m
                                                                                      #  
    cm_flange = ((mass_shoulder * cm_shoulder) \
                  + (mass_tzone1   * cm_tzone1) \
                  + (mass_tzone2   * cm_tzone2) \
                  + (mass_bcflange * cm_bcflange)) \
             / (mass_shoulder + mass_tzone1 + mass_tzone2 + mass_bcflange)            #  R58 Mainshaft Center of Mass  0.529  m

    cost_flange = mass_flange * COST_FORGING                                          #  R64 Total Mainshaft Flange Cost  43,674.50  USD 
	                                                                                  #  R65 Total Rotor Lock Cost    USD
    # TODO: add lockplate - mass, cm, cost
    
    if debug:
        sys.stderr.write('msFlange: len {:.2f} m  mass {:.1f} kg  cm {:.2f} m  cost ${:.2f}\n'.format(flange_length, mass_flange, cm_flange, cost_flange))

    return flange_length, mass_flange, cm_flange, cost_flange

#%% Transform functions for rotor forces and moments (not currently used)
# ---------------------------------------------------------------------
class blade_moment_transform(object):
    ''' Blade_Moment_Transform class          
          The Blade_Moment_Transform class is used to transform moments from the WISDEM rotor models to driveSE.
    '''

    def __init__(self):
        super(blade_moment_transform, self).__init__()

        # variables
        # ensure angles are in radians. Azimuth is 3-element array with blade
        # azimuths; b1, b2, b3 are 3-element arrays for each blade moment (Mx, My,
        # Mz); pitch and cone are floats
        self.add_param('azimuth_angle', val=np.array([0, 2 * pi / 3, 4 * pi / 3]), units='rad', desc='azimuth angles for each blade')
        self.add_param('pitch_angle', val=0.0, units='rad', desc='pitch angle at each blade, assumed same')
        self.add_param('cone_angle', val=0.0, units='rad', desc='cone angle at each blade, assumed same')
        self.add_param('b1', val=np.array([]), units='N*m', desc='moments in x,y,z directions along local blade coordinate system')
        self.add_param('b2', val=np.array([]), units='N*m', desc='moments in x,y,z directions along local blade coordinate system')
        self.add_param('b3', val=np.array([]), units='N*m', desc='moments in x,y,z directions along local blade coordinate system')

        # returns
        self.add_output('Mx', val=0.0, units='N*m',  desc='rotor moment in x-direction')
        self.add_output('My', val=0.0, units='N*m',  desc='rotor moment in y-direction')
        self.add_output('Mz', val=0.0, units='N*m',  desc='rotor moment in z-direction')

    def solve_nonlinear(self, params, unknowns, resids):

        # nested function for transformations
        def trans(alpha, con, phi, bMx, bMy, bMz):
            Mx = bMx * cos(con) * cos(alpha) - bMy * (sin(con) * cos(alpha) * sin(phi) - sin(
                alpha) * cos(phi)) + bMz * (sin(con) * cos(alpha) * cos(phi) - sin(alpha) * sin(phi))
            My = bMx * cos(con) * sin(alpha) - bMy * (sin(con) * sin(alpha) * sin(phi) + cos(
                alpha) * cos(phi)) + bMz * (sin(con) * sin(alpha) * cos(phi) + cos(alpha) * sin(phi))
            Mz = bMx * (-sin(alpha)) - bMy * (-cos(alpha) *
                                              sin(phi)) + bMz * (cos(alpha) * cos(phi))

            return [Mx, My, Mz]

        [b1Mx, b1My, b1Mz] = trans(self.pitch_angle, self.cone_angle, self.azimuth_angle[
                                   0], self.b1[0], self.b1[1], self.b1[2])
        [b2Mx, b2My, b2Mz] = trans(self.pitch_angle, self.cone_angle, self.azimuth_angle[
                                   1], self.b2[0], self.b2[1], self.b2[2])
        [b3Mx, b3My, b3Mz] = trans(self.pitch_angle, self.cone_angle, self.azimuth_angle[
                                   2], self.b3[0], self.b3[1], self.b3[2])

        self.Mx = b1Mx + b2Mx + b3Mx
        self.My = b1My + b2My + b3My
        self.Mz = b1Mz + b2Mz + b3Mz


class blade_force_transform(object):
    ''' Blade_Force_Transform class          
          The Blade_Force_Transform class is used to transform forces from the WISDEM rotor models to driveSE.
    '''

    def __init__(self):
        super(blade_force_transform, self).__init__()

        # variables
        # ensure angles are in radians. Azimuth is 3-element array with blade
        # azimuths; b1, b2, b3 are 3-element arrays for each blade force (Fx, Fy,
        # Fz); pitch and cone are floats
        self.add_param('azimuth_angle', val=np.array([0, 2 * pi / 3, 4 * pi / 3]), units='rad', desc='azimuth angles for each blade')
        self.add_param('pitch_angle', val=0.0, units='rad', desc='pitch angle at each blade, assumed same')
        self.add_param('cone_angle', val=0.0, units='rad', desc='cone angle at each blade, assumed same')
        self.add_param('b1', val=np.array([]), units='N', desc='forces in x,y,z directions along local blade coordinate system')
        self.add_param('b2', val=np.array([]), units='N', desc='forces in x,y,z directions along local blade coordinate system')
        self.add_param('b3', val=np.array([]), units='N', desc='forces in x,y,z directions along local blade coordinate system')

        # returns
        self.add_output('Fx', val=0.0, units='N',  desc='rotor force in x-direction')
        self.add_output('Fy', val=0.0, units='N',  desc='rotor force in y-direction')
        self.add_output('Fz', val=0.0, units='N',  desc='rotor force in z-direction')
        
    def solve_nonlinear(self, params, unknowns, resids):

        # nested function for transformations
        def trans(alpha, con, phi, bFx, bFy, bFz):
            Fx = bFx * cos(con) * cos(alpha) - bFy * (sin(con) * cos(alpha) * sin(phi) - sin(
                alpha) * cos(phi)) + bFz * (sin(con) * cos(alpha) * cos(phi) - sin(alpha) * sin(phi))
            Fy = bFx * cos(con) * sin(alpha) - bFy * (sin(con) * sin(alpha) * sin(phi) + cos(
                alpha) * cos(phi)) + bFz * (sin(con) * sin(alpha) * cos(phi) + cos(alpha) * sin(phi))
            Fz = bFx * (-sin(alpha)) - bFy * (-cos(alpha) *
                                              sin(phi)) + bFz * (cos(alpha) * cos(phi))

            return [Fx, Fy, Fz]

        [b1Fx, b1Fy, b1Fz] = trans(self.pitch_angle, self.cone_angle, self.azimuth_angle[
                                   0], self.b1[0], self.b1[1], self.b1[2])
        [b2Fx, b2Fy, b2Fz] = trans(self.pitch_angle, self.cone_angle, self.azimuth_angle[
                                   1], self.b2[0], self.b2[1], self.b2[2])
        [b3Fx, b3Fy, b3Fz] = trans(self.pitch_angle, self.cone_angle, self.azimuth_angle[
                                   2], self.b3[0], self.b3[1], self.b3[2])

        self.Fx = b1Fx + b2Fx + b3Fx
        self.Fy = b1Fy + b2Fy + b3Fy
        self.Fz = b1Fz + b2Fz + b3Fz

#-------------------------------------------------------------------------
# Fatigue calculations supporting functions and code for low speed shaft and main bearing(s) (not currently used)
# Developed 2014 by Taylor Parsons - requires additional testing and development to complete
#-------------------------------------------------------------------------

'''
# basic supporting functions
def Ninterp(S, a, b):
    return (S / a)**(1 / b)

def Goodman(S_alt, S_mean, Sut):
    return S_alt / (1 - (S_mean / Sut))

def standardrange(N, N_f, Beta, k_b):
    F_delta = (Beta * (log10(N_f) - log10(N))) + 0.18
    if F_delta >= 2 * k_b:
        F_delta = 0.
    return F_delta

# calculate required dynamic load rating, C
def C_calc(F_a, F_r, N_array, p, e, Y1, Y2, X2, life_bearing):
    Fa_ref = np.max(F_a)  # used in comparisons Fa/Fr <e
    Fr_ref = np.max(F_r)

    if Fa_ref / Fr_ref <= e:
        P = F_r + Y1 * F_a
    else:
        P = X2 * F_r + Y2 * F_a

    P_eq = ((scp.integrate.simps((P**p), x=N_array, even='avg')) /
            (N_array[-1] - N_array[0]))**(1 / p)
    C_min = P_eq * (life_bearing / 1e6)**(1. / p) / 1000  # kN
    return C_min

# fatigue analysis for bearings
def fatigue_for_bearings(D_shaft, F_r, F_a, N_array, life_bearing, type, deriv):
    # deriv is boolean, defines if derivatives are returned
    if type == 'CARB':  # p = Fr, so X=1, Y=0
        if (np.max(F_a)) > 0:
            print('---------------------------------------------------------')
            print("error: axial loads too large for CARB bearing application")
            print('---------------------------------------------------------')
        else:
            e = 1
            Y1 = 0.
            X2 = 1.
            Y2 = 0.
            p = 10. / 3
        C_min = C_calc(F_a, F_r, N_array, p, e, Y1, Y2, X2, life_bearing)
        if C_min > 13980 * D_shaft**1.5602:
            out = [D_shaft, 0.4299 * D_shaft +
                   0.0382, 3682.8 * D_shaft**2.7676]
            if deriv:
                out.extend([1., 0.4299, 3682.8 * 2.7676 * D_shaft**1.7676])
        else:
            out = [D_shaft, .2663 * D_shaft + .0435, 1561.4 * D_shaft**2.6007]
            if deriv:
                out.extend([1., .2663, 1561.4 * 2.6007 * D_shaft**1.6007])

    elif type == 'SRB':
        e = 0.32
        Y1 = 2.1
        X2 = 0.67
        Y2 = 3.1
        p = 10. / 3
        C_min = C_calc(F_a, F_r, N_array, p, e, Y1, Y2, X2, life_bearing)
        if C_min > 13878 * D_shaft**1.0796:
            out = [D_shaft, .4801 * D_shaft, 2688.3 * D_shaft**1.8877]
            if deriv:
                out.extend([1., .4801, 2688.3 * 1.8877 * D_shaft**0.8877])
        else:
            out = [D_shaft, .2762 * D_shaft, 876.7 * D_shaft**1.7195]
            if deriv:
                out.extend([1., .2762, 876.7 * 1.7195 * D_shaft**0.7195])

    elif type == 'TRB1':
        e = .37
        Y1 = 0
        X2 = .4
        Y2 = 1.6
        p = 10. / 3
        C_min = C_calc(F_a, F_r, N_array, p, e, Y1, Y2, X2, life_bearing)
        if C_min > 670 * D_shaft + 1690:
            out = [D_shaft, .1335, 269.83 * D_shaft**.441]
            if deriv:
                out.extend([1., 0., 269.83 * 0.441 * D_shaft**(0.441 - 1.)])
        else:
            out = [D_shaft, .0740, 92.863 * D_shaft**.8399]
            if deriv:
                out.extend([1., 0., 92.863 * 0.8399 * D_shaft**(0.8399 - 1.)])

    elif type == 'CRB':
        if (np.max(F_a) / np.max(F_r) >= .5) or (np.min(F_a) / (np.min(F_r)) >= .5):
            print('--------------------------------------------------------')
            print("error: axial loads too large for CRB bearing application")
            print('--------------------------------------------------------')
        else:
            e = 0.2
            Y1 = 0
            X2 = 0.92
            Y2 = 0.6
            p = 10. / 3
            C_min = C_calc(F_a, F_r, N_array, p, e, Y1, Y2, X2, life_bearing)
            if C_min > 4526.5 * D_shaft**.9556:
                out = [D_shaft, .2603 * D_shaft, 1070.8 * D_shaft**1.8278]
                if deriv:
                    out.extend([1., .2603, 1070.8 * 1.8278 * D_shaft**0.8278])
            else:
                out = [D_shaft, .1136 * D_shaft, 304.19 * D_shaft**1.8885]
                if deriv:
                    out.extend([1., .1136, 304.19 * 1.8885 * D_shaft**0.8885])

    elif type == 'TRB2':
        e = 0.4
        Y1 = 2.5
        X2 = 0.4
        Y2 = 1.75
        p = 10. / 3
        C_min = C_calc(F_a, F_r, N_array, p, e, Y1, Y2, X2, life_bearing)
        if C_min > 6579.9 * D_shaft**.8592:
            out = [D_shaft, .3689 * D_shaft, 1442.6 * D_shaft**1.8932]
            if deriv:
                out.extend([1., .3689, 1442.6 * 1.8932 * D_shaft**.8932])
        else:
            out = [D_shaft, .1499 * D_shaft, 543.01 * D_shaft**1.9043]
            if deriv:
                out.extend([1., .1499, 543.01 * 1.9043 * D_shaft**.9043])

    elif type == 'RB':  # factors depend on ratio Fa/C0, C0 depends on bearing... TODO: add this functionality
        e = 0.4
        Y1 = 1.6
        X2 = 0.75
        Y2 = 2.15
        p = 3.
        C_min = C_calc(F_a, F_r, N_array, p, e, Y1, Y2, X2, life_bearing)
        if C_min > 884.5 * D_shaft**.9964:
            out = [D_shaft, .1571, 646.46 * D_shaft**2.]
            if deriv:
                out.extend([1., 0., 646.46 * 2. * D_shaft])
        else:
            out = [D_shaft, .0839, 229.47 * D_shaft**1.8036]
            if deriv:
                out.extend([1.0, 0.0, 229.47 * 1.8036 * D_shaft**0.8036])

    return out

#-------------------------------------------------------------------------
def get_Damage_Brng2(self):
    I = (pi / 64.0) * (self.D_med**4 - self.D_in**4)
    J = I * 2
    Area = pi / 4. * (self.D_med**2 - self.D_in**2)
    self.LssWeight = self.density * 9.81 * \
        (((pi / 12) * (self.D_max**2 + self.D_med**2 + self.D_max *
                       self.D_med) * (self.L_mb)) - (pi / 4 * self.L_mb * self.D_in**2))

    self.Fz1stoch = (-self.My_stoch) / (self.L_mb)
    self.Fy1stoch = self.Mz_stoch / self.L_mb
    self.My2stoch = 0.  # My_stoch - abs(Fz1stoch)*self.L_mb #=0
    self.Mz2stoch = 0.  # Mz_stoch - abs(Fy1stoch)*self.L_mb #=0

    # create stochastic loads across N
    stoch_bend2 = (self.My2stoch**2 + self.Mz2stoch **
                   2)**(0.5) * self.D_med / (2. * I)
    stoch_shear2 = abs(self.Mx_stoch * self.D_med / (2. * J))
    # all normal force held by downwind bearing
    stoch_normal2 = self.Fx_stoch / Area * cos(self.shaft_angle)
    stoch_stress2 = ((stoch_bend2 + stoch_normal2) **
                     2 + 3. * stoch_shear2**2)**(0.5)

    # create mean loads
    # Fz_mean*self.distance_hub2mb*self.D_med/(2.*I) #not mean, but deterministic
    mean_bend2 = 0.
    mean_shear2 = self.Mx_mean * self.D_med / (2. * J)
    mean_normal2 = self.Fx_mean / Area * \
        cos(self.shaft_angle) + (self.rotorWeight +
                                 self.LssWeight) * sin(self.shaft_angle)
    mean_stress2 = ((mean_bend2 + mean_normal2) **
                    2 + 3. * mean_shear2**2)**(0.5)
    # apply Goodman with compressive (-) mean stress
    S_mod_stoch2 = Goodman(stoch_stress2, -mean_stress2, self.S_ut)

    # Use Palmgren-Miner linear damage rule to add damage from stochastic
    # load ranges
    DEL_y = self.Fx_stoch.copy()  # initialize
    for i in range(self.num_pts):
        DEL_y[i] = self.N[i] / \
            (Ninterp(S_mod_stoch2[i], self.SN_a, self.SN_b))

    # damage from stochastic loading
    self.Damage = scp.integrate.simps(DEL_y, x=self.N, even='avg')

    # create deterministic loads occurring N_rotor times
    self.Fz1determ = (self.gearboxWeight * self.L_gb - self.LssWeight * .5 *
                      self.L_mb - self.rotorWeight * (self.L_mb + self.distance_hub2mb)) / (self.L_mb)
    # -rotorWeight*(self.distance_hub2mb+self.L_mb) + Fz1determ*self.L_mb - self.LssWeight*.5*self.L_mb + self.gearboxWeight*self.L_gb
    self.My2determ = self.gearboxWeight * self.L_gb
    self.determ_stress2 = abs(self.My2determ * self.D_med / (2. * I))

    S_mod_determ2 = Goodman(self.determ_stress2, -mean_stress2, self.S_ut)

    if S_mod_determ2 > 0:
        self.Damage += self.N_rotor / \
            (Ninterp(S_mod_determ2, self.SN_a, self.SN_b))

def get_Damage_Brng1(self):
    self.D_in = self.shaft_ratio * self.D_max
    self.D_max = (self.D_max**4 + self.D_in**4)**0.25
    self.D_min = (self.D_min**4 + self.D_in**4)**0.25
    I = (pi / 64.0) * (self.D_max**4 - self.D_in**4)
    J = I * 2
    Area = pi / 4. * (self.D_max**2 - self.D_in**2)
    self.LssWeight = self.density * 9.81 * \
        (((pi / 12) * (self.D_max**2 + self.D_min**2 + self.D_max *
                       self.D_min) * (self.L_ms)) - (pi / 4 * self.L_ms * self.D_in**2))

    # create stochastic loads across N
    stoch_bend1 = (self.My_stoch**2 + self.Mz_stoch **
                   2)**(0.5) * self.D_max / (2. * I)
    stoch_shear1 = abs(self.Mx_stoch * self.D_max / (2. * J))
    stoch_normal1 = self.Fx_stoch / Area * cos(self.shaft_angle)
    stoch_stress1 = ((stoch_bend1 + stoch_normal1) **
                     2 + 3. * stoch_shear1**2)**(0.5)

    # create mean loads
    # Fz_mean*self.distance_hub2mb*self.D_max/(2.*I) #not mean, but deterministic
    mean_bend1 = 0
    mean_shear1 = self.Mx_mean * self.D_max / (2. * J)
    mean_normal1 = self.Fx_mean / Area * \
        cos(self.shaft_angle) + (self.rotorWeight +
                                 self.LssWeight) * sin(self.shaft_angle)
    mean_stress1 = ((mean_bend1 + mean_normal1) **
                    2 + 3. * mean_shear1**2)**(0.5)

    # apply Goodman with compressive (-) mean stress
    S_mod_stoch1 = Goodman(stoch_stress1, -mean_stress1, self.S_ut)

    # Use Palmgren-Miner linear damage rule to add damage from stochastic
    # load ranges
    DEL_y = self.Fx_stoch.copy()  # initialize
    for i in range(self.num_pts):
        DEL_y[i] = self.N[i] / \
            (Ninterp(S_mod_stoch1[i], self.SN_a, self.SN_b))

    # damage from stochastic loading
    self.Damage = scp.integrate.simps(DEL_y, x=self.N, even='avg')

    # create deterministic loads occurring N_rotor times
    # only deterministic stress at mb1 is bending due to weights
    determ_stress1 = abs(
        self.rotorWeight * cos(self.shaft_angle) * self.distance_hub2mb * self.D_max / (2. * I))

    S_mod_determ = Goodman(determ_stress1, -mean_stress1, self.S_ut)

    self.Damage += self.N_rotor / \
        (Ninterp(S_mod_determ, self.SN_a, self.SN_b))

def setup_Fatigue_Loads(self):
    R = self.rotor_diameter / 2.0
    rotor_torque = (self.machine_rating * 1000 /
                    self.drivetrain_efficiency) / (self.rotor_freq * (pi / 30))
    Tip_speed_ratio = self.rotor_freq / 30. * pi * R / self.Vrated
    rho_air = 1.225  # kg/m^3 density of air TODO add as input
    p_o = 4. / 3 * rho_air * ((4 * pi * self.rotor_freq / 60 * R / 3)**2 + self.Vrated**2) * (
        pi * R / (self.blade_number * Tip_speed_ratio * (Tip_speed_ratio**2 + 1)**(.5)))
    # characteristic frequency on rotor from turbine of given blade number
    # [Hz]
    n_c = self.blade_number * self.rotor_freq / 60
    # number of rotor rotations based off of weibull curve. .827 comes from
    # lower rpm than rated at lower wind speeds
    self.N_f = self.availability * n_c * (self.T_life * 365 * 24 * 60 * 60) * exp(-(
        self.cut_in / self.weibull_A)**self.weibull_k) - exp(-(self.cut_out / self.weibull_A)**self.weibull_k)

    k_b = 2.5  # calculating rotor pressure from all three blades. Use kb=1 for individual blades

    if self.IEC_Class == 'A':  # From IEC 61400-1 TODO consider calculating based off of 10-minute windspeed and weibull parameters, include neighboring wake effects?
        I_t = 0.18
    elif self.IEC_Class == 'B':
        I_t = 0.14
    else:
        I_t = 0.12

    Beta = 0.11 * k_b * (I_t + 0.1) * (self.weibull_A + 4.4)

    # for analysis with N on log scale, makes larger loads contain finer
    # step sizes
    self.num_pts = 100
    # with zeros: N=np.logspace(log10(1.0),log10(N_f),endpoint=True,num=self.num_pts)
    self.N = np.logspace((log10(self.N_f) - (2 * k_b - 0.18) / Beta),
                         log10(self.N_f), endpoint=True, num=self.num_pts)
    self.N_rotor = self.N_f / 3.
    F_stoch = self.N.copy()

    k_r = 0.8  # assuming natural frequency of rotor is significantly larger than rotor rotational frequency

    for i in range(self.num_pts):
        F_stoch[i] = self.standardrange(self.N[i], self.N_f, Beta, k_b)

    Fx_factor = (.3649 * log(self.rotor_diameter) - 1.074)
    Mx_factor = (.0799 * log(self.rotor_diameter) - .2577)
    My_factor = (.172 * log(self.rotor_diameter) - .5943)
    Mz_factor = (.1659 * log(self.rotor_diameter) - .5795)

    self.Fx_stoch = (F_stoch.copy() * 0.5 * p_o * (R)) * Fx_factor
    self.Mx_stoch = (F_stoch.copy() * 0.45 * p_o *
                     (R)**2) * Mx_factor  # *0.31
    self.My_stoch = (F_stoch.copy() * 0.33 * p_o *
                     k_r * (R)**2) * My_factor  # *0.25
    self.Mz_stoch = (F_stoch.copy() * 0.33 * p_o *
                     k_r * (R)**2) * Mz_factor  # *0.25

    self.Fx_mean = 0.5 * p_o * R * self.blade_number * Fx_factor
    self.Mx_mean = 0.5 * rotor_torque * Mx_factor
    self.rotorWeight = self.rotor_mass * self.g


# Code remove from LowSpeedShaft4pt component
############################################
# inputs into LSS for fatigue calculations:

check_fatigue = Enum(0,(0,1,2),iotype = 'in', desc = 'turns on and off fatigue check')

#fatigue1 variables
rotor_freq = Float(iotype = 'in', units = 'rpm', desc='rated rotor speed')
availability = Float(.95,iotype = 'in', desc = 'turbine availability')
fatigue_exponent = Float(0,iotype = 'in', desc = 'fatigue exponent of material')
S_ut = Float(700e6,iotype = 'in', units = 'Pa', desc = 'ultimate tensile strength of material')
weibull_A = Float(iotype = 'in', units = 'm/s', desc = 'weibull scale parameter "A" of 10-minute windspeed probability distribution')
weibull_k = Float(iotype = 'in', desc = 'weibull shape parameter "k" of 10-minute windspeed probability distribution')
blade_number = Float(iotype = 'in', desc = 'number of blades on rotor, 2 or 3')
cut_in = Float(iotype = 'in', units = 'm/s', desc = 'cut-in windspeed')
cut_out = Float(iotype = 'in', units = 'm/s', desc = 'cut-out windspeed')
Vrated = Float(iotype = 'in', units = 'm/s', desc = 'rated windspeed')
T_life = Float(iotype = 'in', units = 'yr', desc = 'cut-in windspeed')

# fatigue2 variables
rotor_thrust_distribution = Array(iotype='in', units ='N', desc = 'thrust distribution across turbine life')
rotor_thrust_count = Array(iotype='in', desc = 'corresponding cycle array for thrust distribution')
rotor_Fy_distribution = Array(iotype='in', units ='N', desc = 'Fy distribution across turbine life')
rotor_Fy_count = Array(iotype='in', desc = 'corresponding cycle array for Fy distribution')
rotor_Fz_distribution = Array(iotype='in', units ='N', desc = 'Fz distribution across turbine life')
rotor_Fz_count = Array(iotype='in', desc = 'corresponding cycle array for Fz distribution') 
rotor_torque_distribution = Array(iotype='in', units ='N*m', desc = 'torque distribution across turbine life')
rotor_torque_count = Array(iotype='in', desc = 'corresponding cycle array for torque distribution') 
rotor_My_distribution = Array(iotype='in', units ='N*m', desc = 'My distribution across turbine life')
rotor_My_count = Array(iotype='in', desc = 'corresponding cycle array for My distribution') 
rotor_Mz_distribution = Array(iotype='in', units ='N*m', desc = 'Mz distribution across turbine life')
rotor_Mz_count = Array(iotype='in', desc = 'corresponding cycle array for Mz distribution')

# LSS fatigue check calculations 
# fatigue check Taylor Parsons 6/14
if self.check_fatigue == 1 or self.check_fatigue == 2:

    # checks to make sure all inputs are reasonable
    if self.rotor_mass < 100:
        [self.rotor_mass] = get_rotor_mass(self.machine_rating, False)

    # material properties 34CrNiMo6 steel +QT, large diameter
    self.n_safety = 2.5
    if self.S_ut <= 0:
        self.S_ut = 700.0e6  # Pa

    # calculate material props for fatigue
    Sm = 0.9 * self.S_ut  # for bending situations, material strength at 10^3 cycles

    if self.fatigue_exponent != 0:
        if self.fatigue_exponent > 0:
            self.SN_b = - self.fatigue_exponent
        else:
            self.SN_b = self.fatigue_exponent
    else:
        C_size = 0.6  # diameter larger than 10"
        # machined surface 272*(self.S_ut/1e6)**-.995 #forged
        C_surf = 4.51 * (self.S_ut / 1e6)**-.265
        C_temp = 1  # normal operating temps
        C_reliab = 0.814  # 99% reliability
        C_envir = 1.  # enclosed environment
        Se = C_size * C_surf * C_temp * C_reliab * C_envir * .5 * \
            self.S_ut  # modified endurance limit for infinite life (should be Sf)\
        Nfinal = 5e8  # point where fatigue limit occurs under hypothetical S-N curve TODO adjust to fit actual data
        # assuming no endurance limit (high strength steel)
        z = log10(1e3) - log10(Nfinal)
        self.SN_b = 1 / z * log10(Sm / Se)
    self.SN_a = Sm / (1000.**self.SN_b)


    if self.check_fatigue == 1:

        setup_Fatigue_Loads(self)

        # upwind bearing calculations
        iterationstep = 0.001
        diameter_limit = 5.0
        while True:
            get_Damage_Brng1(self)
            if self.Damage < 1 or self.D_max >= diameter_limit:
                break
            else:
                self.D_max += iterationstep

        # downwind bearing calculations
        diameter_limit = 5.0
        iterationstep = 0.001
        while True:
            get_Damage_Brng2(self)
            if self.Damage < 1 or self.D_med >= diameter_limit:
                break
            else:
                self.D_med += iterationstep

        # begin bearing calculations
        # counts per rotation (not defined by characteristic frequency
        # 3n_rotor)
        N_bearings = self.N / self.blade_number

        # radial stochastic + deterministic mean
        Fr1_range = ((abs(self.Fz1stoch) + abs(self.Fz1determ))
                     ** 2 + self.Fy1stoch**2)**.5
        Fa1_range = np.zeros(len(self.Fy1stoch))

        #...calculate downwind forces
        lss_weight = self.density * 9.81 * \
            (((pi / 12) * (self.D_max**2 + self.D_med**2 + self.D_max *
                           self.D_med) * (self.L_mb)) - (pi / 4 * self.L_mb * self.D_in**2))
        Fy2stoch = -self.Mz_stoch / (self.L_mb)  # = -Fy1 - Fy_stoch
        Fz2stoch = -(lss_weight * 2. / 3. * self.L_mb - self.My_stoch) / (self.L_mb) + (lss_weight + self.shrinkDiscWeight + self.gearboxWeight) * \
            cos(self.shaft_angle) - \
            self.rotorWeight  # -Fz1 +Weights*cos(gamma)-Fz_stoch+Fz_mean (Fz_mean is in negative direction)
        Fr2_range = (Fy2stoch**2 + (Fz2stoch + abs(-self.rotorWeight * distance_hub2mb +
                                                   0.5 * lss_weight + self.gearboxWeight * self.L_gb / self.L_mb))**2)**0.5
        Fa2_range = self.Fx_stoch * cos(self.shaft_angle) + (
            self.rotorWeight + lss_weight) * sin(self.shaft_angle)  # axial stochastic + mean

        life_bearing = self.N_f / self.blade_number

        [self.D_max_a, facewidth_max, bearing1mass] = fatigue_for_bearings(
            self.D_max, Fr1_range, Fa1_range, N_bearings, life_bearing, self.mb1Type, False)
        [self.D_med_a, facewidth_med, bearing2mass] = fatigue_for_bearings(
            self.D_med, Fr2_range, Fa2_range, N_bearings, life_bearing, self.mb2Type, False)

    # elif self.check_fatigue == 2: # untested and not used currently
    #   Fx = self.rotor_thrust_distribution
    #   n_Fx = self.rotor_thrust_count
    #   Fy = self.rotor_Fy_distribution
    #   n_Fy = self.rotor_Fy_count
    #   Fz = self.rotor_Fz_distribution
    #   n_Fz = self.rotor_Fz_count
    #   Mx = self.rotor_torque_distribution
    #   n_Mx = self.rotor_torque_count
    #   My = self.rotor_My_distribution
    #   n_My = self.rotor_My_count
    #   Mz = self.rotor_Mz_distribution
    #   n_Mz = self.rotor_Mz_count

    #   def Ninterp(L_ult,L_range,m):
    # return (L_ult/(.5*L_range))**m #TODO double-check that the input
    # will be the load RANGE instead of load amplitudes. May also
    # include means

    #   #upwind bearing calcs
    #   diameter_limit = 5.0
    #   iterationstep=0.001
    #   #upwind bearing calcs
    #   while True:
    #       self.Damage = 0
    #       Fx_ult = self.SN_a*(pi/4.*(self.D_max**2-self.D_in**2))
    #       Fyz_ult = self.SN_a*(pi*(self.D_max**4-self.D_in**4))/(self.D_max*64.)/self.distance_hub2mb
    #       Mx_ult = self.SN_a*(pi*(self.D_max**4-self.D_in**4))/(32*(3)**.5*self.D_max)
    #       Myz_ult = self.SN_a*(pi*(self.D_max**4-self.D_in**4))/(self.D_max*64.)
    #       if Fx_ult !=0 and np.all(n_Fx) != 0:
    #           self.Damage+=scp.integrate.simps(n_Fx/Ninterp(Fx_ult,Fx,-1/self.SN_b),x=n_Fx,even = 'avg')
    #       if Fyz_ult !=0:
    #           if np.all(n_Fy) != 0:
    #               self.Damage+=scp.integrate.simps(abs(n_Fy/Ninterp(Fyz_ult,Fy,-1/self.SN_b)),x=n_Fy,even = 'avg')
    #           if np.all(n_Fz) != 0:
    #               self.Damage+=scp.integrate.simps(abs(n_Fz/Ninterp(Fyz_ult,Fz,-1/self.SN_b)),x=n_Fz,even = 'avg')
    #       if Mx_ult !=0 and np.all(n_Mx) != 0:
    #           self.Damage+=scp.integrate.simps(abs(n_Mx/Ninterp(Mx_ult,Mx,-1/self.SN_b)),x=n_Mx,even = 'avg')
    #       if Myz_ult!=0:
    #           if np.all(n_My) != 0:
    #               self.Damage+=scp.integrate.simps(abs(n_My/Ninterp(Myz_ult,My,-1/self.SN_b)),x=n_My,even = 'avg')
    #           if np.all(n_Mz) != 0:
    #               self.Damage+=scp.integrate.simps(abs(n_Mz/Ninterp(Myz_ult,Mz,-1/self.SN_b)),x=n_Mz,even = 'avg')

    #       if self.Damage <= 1 or self.D_max >= diameter_limit:
    #           break
    #       else:
    #           self.D_max+=iterationstep
    #   #downwind bearing calcs
    #   while True:
    #       self.Damage = 0
    #       Fx_ult = self.SN_a*(pi/4.*(self.D_med**2-self.D_in**2))
    #       Mx_ult = self.SN_a*(pi*(self.D_med**4-self.D_in**4))/(32*(3)**.5*self.D_med)
    #       if Fx_ult !=0:
    #           self.Damage+=scp.integrate.simps(n_Fx/Ninterp(Fx_ult,Fx,-1/self.SN_b),x=n_Fx,even = 'avg')
    #       if Mx_ult !=0:
    #           self.Damage+=scp.integrate.simps(n_Mx/Ninterp(Mx_ult,Mx,-1/self.SN_b),x=n_Mx,even = 'avg')

    #       if self.Damage <= 1 or self.D_med>= diameter_limit:
    #           break
    #       else:
    #           self.D_med+=iterationstep

    #   #bearing calcs
    #   if self.availability != 0 and rotor_freq != 0 and self.T_life != 0 and self.cut_out != 0 and self.weibull_A != 0:
    #       N_rotations = self.availability*rotor_freq/60.*(self.T_life*365*24*60*60)*exp(-(self.cut_in/self.weibull_A)**self.weibull_k)-exp(-(self.cut_out/self.weibull_A)**self.weibull_k)
    #   elif np.max(n_Fx > 1e6):
    #       N_rotations = np.max(n_Fx)/self.blade_number
    #   elif np.max(n_My > 1e6):
    #       N_rotations = np.max(n_My)/self.blade_number
    #   Fz1_Fz = Fz*(self.L_mb+self.distance_hub2mb)/self.L_mb
    #   Fz1_My = My/self.L_mb
    #   Fy1_Fy = -Fy*(self.L_mb+self.distance_hub2mb)/self.L_mb
    #   Fy1_Mz = Mz/self.L_mb
    #   [self.D_max_a,facewidth_max,bearing1mass] = fatigue2_for_bearings(self.D_max,self.mb1Type,np.zeros(2),np.array([1,2]),Fy1_Fy,n_Fy/self.blade_number,Fz1_Fz,n_Fz/self.blade_number,Fz1_My,n_My/self.blade_number,Fy1_Mz,n_Mz/self.blade_number,N_rotations)
    #   Fz2_Fz = Fz*self.distance_hub2mb/self.L_mb
    #   Fz2_My = My/self.L_mb
    #   Fy2_Fy = Fy*self.distance_hub2mb/self.L_mb
    #   Fy2_Mz = Mz/self.L_mb
    #   [self.D_med_a,facewidth_med,bearing2mass] = fatigue2_for_bearings(self.D_med,self.mb2Type,Fx,n_Fx/self.blade_number,Fy2_Fy,n_Fy/self.blade_number,Fz2_Fz,n_Fz/self.blade_number,Fz2_My,n_My/self.blade_number,Fy2_Mz,n_Mz/self.blade_number,N_rotations)


# Code remove from LowSpeedShaft3pt component
############################################
# inputs into LSS for fatigue calculations:

check_fatigue = Enum(0,(0,1,2),iotype = 'in', desc = 'turns on and off fatigue check')

#fatigue1 variables
rotor_freq = Float(iotype = 'in', units = 'rpm', desc='rated rotor speed')
availability = Float(.95,iotype = 'in', desc = 'turbine availability')
fatigue_exponent = Float(0,iotype = 'in', desc = 'fatigue exponent of material')
S_ut = Float(700e6,iotype = 'in', units = 'Pa', desc = 'ultimate tensile strength of material')
weibull_A = Float(iotype = 'in', units = 'm/s', desc = 'weibull scale parameter "A" of 10-minute windspeed probability distribution')
weibull_k = Float(iotype = 'in', desc = 'weibull shape parameter "k" of 10-minute windspeed probability distribution')
blade_number = Float(iotype = 'in', desc = 'number of blades on rotor, 2 or 3')
cut_in = Float(iotype = 'in', units = 'm/s', desc = 'cut-in windspeed')
cut_out = Float(iotype = 'in', units = 'm/s', desc = 'cut-out windspeed')
Vrated = Float(iotype = 'in', units = 'm/s', desc = 'rated windspeed')
T_life = Float(iotype = 'in', units = 'yr', desc = 'cut-in windspeed')

# fatigue2 variables
rotor_thrust_distribution = Array(iotype='in', units ='N', desc = 'thrust distribution across turbine life')
rotor_thrust_count = Array(iotype='in', desc = 'corresponding cycle array for thrust distribution')
rotor_Fy_distribution = Array(iotype='in', units ='N', desc = 'Fy distribution across turbine life')
rotor_Fy_count = Array(iotype='in', desc = 'corresponding cycle array for Fy distribution')
rotor_Fz_distribution = Array(iotype='in', units ='N', desc = 'Fz distribution across turbine life')
rotor_Fz_count = Array(iotype='in', desc = 'corresponding cycle array for Fz distribution') 
rotor_torque_distribution = Array(iotype='in', units ='N*m', desc = 'torque distribution across turbine life')
rotor_torque_count = Array(iotype='in', desc = 'corresponding cycle array for torque distribution') 
rotor_My_distribution = Array(iotype='in', units ='N*m', desc = 'My distribution across turbine life')
rotor_My_count = Array(iotype='in', desc = 'corresponding cycle array for My distribution') 
rotor_Mz_distribution = Array(iotype='in', units ='N*m', desc = 'Mz distribution across turbine life')
rotor_Mz_count = Array(iotype='in', desc = 'corresponding cycle array for Mz distribution')

# fatigue check Taylor Parsons 6/2014
if self.check_fatigue == 1 or 2:
	  #start_time = time.time()
	  #material properties 34CrNiMo6 steel +QT, large diameter
	  self.E=2.1e11
	  self.density=7800.0
	  self.n_safety = 2.5
	  if self.S_ut <= 0:
	    self.S_ut=700.0e6 #Pa
	  Sm=0.9*self.S_ut #for bending situations, material strength at 10^3 cycles
	  C_size=0.6 #diameter larger than 10"
	  C_surf=4.51*(self.S_ut/1e6)**-.265 #machined surface 272*(self.S_ut/1e6)**-.995 #forged
	  C_temp=1 #normal operating temps
	  C_reliab=0.814 #99% reliability
	  C_envir=1. #enclosed environment
	  Se=C_size*C_surf*C_temp*C_reliab*C_envir*.5*self.S_ut #modified endurance limit for infinite life
	
	  if self.fatigue_exponent!=0:
	    if self.fatigue_exponent > 0:
	        self.SN_b = - self.fatigue_exponent
	    else:
	        self.SN_b = self.fatigue_exponent
	  else:
	    Nfinal = 5e8 #point where fatigue limit occurs under hypothetical S-N curve TODO adjust to fit actual data
	    z=log10(1e3)-log10(Nfinal)  #assuming no endurance limit (high strength steel)
	    self.SN_b=1/z*log10(Sm/Se)
	  self.SN_a=Sm/(1000.**self.SN_b)
	  # print 'm:', -1/self.SN_b
	  # print 'a:', self.SN_a
	
	  if self.check_fatigue == 1:
	      #checks to make sure all inputs are reasonable
	      if self.rotor_mass < 100:
	          [self.rotor_mass] = get_rotor_mass(self.machine_rating,False)
	
	      #Rotor Loads calculations using DS472
	      setup_Fatigue_Loads(self)
	
	      #upwind diameter calculations
	      iterationstep=0.001
	      diameter_limit = 1.5
	      while True:
	
	          get_Damage_Brng1(self)
	
	          if self.Damage < 1 or self.D_max >= diameter_limit:
	              break
	          else:
	              self.D_max+=iterationstep
	
	      #begin bearing calculations
	      N_bearings = self.N/self.blade_number #rotation number
	
	      Fz1stoch = (-self.My_stoch)/(self.L_ms)
	      Fy1stoch = self.Mz_stoch/self.L_ms
	      Fz1determ = (self.weightGbx*self.L_gb - self.LssWeight*.5*self.L_ms - self.rotorWeight*(self.L_ms+distance_hub2mb)) / (self.L_ms)
	
	      Fr_range = ((abs(Fz1stoch)+abs(Fz1determ))**2 +Fy1stoch**2)**.5 #radial stochastic + deterministic mean
	      Fa_range = self.Fx_stoch*cos(self.shaft_angle) + (self.rotorWeight+self.LssWeight)*sin(self.shaft_angle) #axial stochastic + mean
	
	      life_bearing = self.N_f/self.blade_number
	
	      [self.D_max_a,facewidth_max,bearingmass] = fatigue_for_bearings(self.D_max, Fr_range, Fa_range, N_bearings, life_bearing, self.mb1Type,False)
	
	  # elif self.check_fatigue == 2:
	  #   Fx = self.rotor_thrust_distribution
	  #   n_Fx = self.rotor_thrust_count
	  #   Fy = self.rotor_Fy_distribution
	  #   n_Fy = self.rotor_Fy_count
	  #   Fz = self.rotor_Fz_distribution
	  #   n_Fz = self.rotor_Fz_count
	  #   Mx = self.rotor_torque_distribution
	  #   n_Mx = self.rotor_torque_count
	  #   My = self.rotor_My_distribution
	  #   n_My = self.rotor_My_count
	  #   Mz = self.rotor_Mz_distribution
	  #   n_Mz = self.rotor_Mz_count
	
	  #   # print n_Fx
	  #   # print Fx*.5
	  #   # print Mx*.5
	  #   # print -1/self.SN_b
	
	  #   def Ninterp(L_ult,L_range,m):
	  #       return (L_ult/(.5*L_range))**m #TODO double-check that the input will be the load RANGE instead of load amplitudes. Also, may include means?
	
	  #   #upwind bearing calcs
	  #   diameter_limit = 5.0
	  #   iterationstep=0.001
	  #   #upwind bearing calcs
	  #   while True:
	  #       self.Damage = 0
	  #       Fx_ult = self.SN_a*(pi/4.*(self.D_max**2-self.D_in**2))
	  #       Fyz_ult = self.SN_a*(pi*(self.D_max**4-self.D_in**4))/(self.D_max*32*self.distance_hub2mb)
	  #       Mx_ult = self.SN_a*(pi*(self.D_max**4-self.D_in**4))/(32*(3.**.5)*self.D_max)
	  #       Myz_ult = self.SN_a*(pi*(self.D_max**4-self.D_in**4))/(self.D_max*64.)
	  #       if Fx_ult and np.all(n_Fx):
	  #           self.Damage+=scp.integrate.simps(n_Fx/Ninterp(Fx_ult,Fx,-1/self.SN_b),x=n_Fx,even = 'avg')
	  #       if Fyz_ult:
	  #           if np.all(n_Fy):
	  #               self.Damage+=scp.integrate.simps(abs(n_Fy/Ninterp(Fyz_ult,Fy,-1/self.SN_b)),x=n_Fy,even = 'avg')
	  #           if np.all(n_Fz):
	  #               self.Damage+=scp.integrate.simps(abs(n_Fz/Ninterp(Fyz_ult,Fz,-1/self.SN_b)),x=n_Fz,even = 'avg')
	  #       if Mx_ult and np.all(n_Mx):
	  #           self.Damage+=scp.integrate.simps(abs(n_Mx/Ninterp(Mx_ult,Mx,-1/self.SN_b)),x=n_Mx,even = 'avg')
	  #       if Myz_ult:
	  #           if np.all(n_My):
	  #               self.Damage+=scp.integrate.simps(abs(n_My/Ninterp(Myz_ult,My,-1/self.SN_b)),x=n_My,even = 'avg')
	  #           if np.all(n_Mz):
	  #               self.Damage+=scp.integrate.simps(abs(n_Mz/Ninterp(Myz_ult,Mz,-1/self.SN_b)),x=n_Mz,even = 'avg')
	
	  #       print 'Upwind Bearing Diameter:', self.D_max
	  #       print 'self.Damage:', self.Damage
	
	  #       if self.Damage <= 1 or self.D_max >= diameter_limit:
	  #           # print 'Upwind Bearing Diameter:', self.D_max
	  #           # print 'self.Damage:', self.Damage
	  #           #print (time.time() - start_time), 'seconds of total simulation time'
	  #           break
	  #       else:
	  #           self.D_max+=iterationstep
	
	  #   #bearing calcs
	  #   if self.availability != 0 and rotor_freq != 0 and self.T_life != 0 and self.cut_out != 0 and self.weibull_A != 0:
	  #       N_rotations = self.availability*rotor_freq/60.*(self.T_life*365*24*60*60)*exp(-(self.cut_in/self.weibull_A)**self.weibull_k)-exp(-(self.cut_out/self.weibull_A)**self.weibull_k)
	  #   elif np.max(n_Fx > 1e6):
	  #       N_rotations = np.max(n_Fx)/self.blade_number
	  #   elif np.max(n_My > 1e6):
	  #       N_rotations = np.max(n_My)/self.blade_number
	
	  #   # Fz1 = (Fz*(self.L_ms+self.distance_hub2mb)+My)/self.L_ms
	  #   Fz1_Fz = Fz*(self.L_ms+self.distance_hub2mb)/self.L_ms #force in z direction due to Fz
	  #   Fz1_My = My/self.L_ms #force in z direction due to My
	  #   Fy1_Fy = -Fy*(self.L_ms+self.distance_hub2mb)/self.L_ms
	  #   Fy1_Mz = Mz/self.L_ms
	  #   [self.D_max_a,facewidth_max,bearingmass] = fatigue2_for_bearings(self.D_max,self.mb1Type,np.zeros(2),np.array([1,2]),Fy1_Fy,n_Fy/self.blade_number,Fz1_Fz,n_Fz/self.blade_number,Fz1_My,n_My/self.blade_number,Fy1_Mz,n_Mz/self.blade_number,N_rotations)
 
'''


import numpy as np
import numpy.testing as npt
import unittest
from openmdao.api import Problem
from wisdem.floatingse.floating import FloatingSE

nSec = 4
nSecTow = 3
NPTS = 100

class TestRegression(unittest.TestCase):
    def setUp(self):
        self.myfloat = Problem()
        self.myfloat.model=FloatingSE()
        self.myfloat.setup()

        # Remove all offset columns
        self.myfloat['number_of_offset_columns'] = 0
        self.myfloat['cross_attachment_pontoons_int']   = 0
        self.myfloat['lower_attachment_pontoons_int']   = 0
        self.myfloat['upper_attachment_pontoons_int']   = 0
        self.myfloat['lower_ring_pontoons_int']         = 0
        self.myfloat['upper_ring_pontoons_int']         = 0
        self.myfloat['outer_cross_pontoons_int']        = 0

        # Wind and water properties
        self.myfloat['air_density'] = 1.226   # Density of air [kg/m^3]
        self.myfloat['air_viscosity']  = 1.78e-5 # Viscosity of air [kg/m/s]
        self.myfloat['water_density']      = 1025.0  # Density of water [kg/m^3]
        self.myfloat['water_viscosity']  = 1.08e-3 # Viscosity of water [kg/m/s]
        self.myfloat['shearExp']    = 0.11   # Shear exponent in wind power law
        self.myfloat['cm']          = 2.0    # Added mass coefficient
        self.myfloat['Uc']          = 0.0    # Mean current speed
        self.myfloat['wind_z0']          = 0.0    # Water line
        self.myfloat['yaw']         = 0.0    # Turbine yaw angle
        self.myfloat['beta']        = 0.0    # Wind beta angle
        self.myfloat['cd_usr']      = -1.0 # Compute drag coefficient

        # Material properties
        self.myfloat['material_density'] = self.myfloat['main.material_density'] = self.myfloat['off.material_density'] = 7850.0          # Steel [kg/m^3]
        self.myfloat['E']                = 200e9           # Young's modulus [N/m^2]
        self.myfloat['G']                = 79.3e9          # Shear modulus [N/m^2]
        self.myfloat['yield_stress']     = 3.45e8          # Elastic yield stress [N/m^2]
        self.myfloat['nu']               = 0.26            # Poisson's ratio
        self.myfloat['permanent_ballast_density'] = 5000.0 # [kg/m^3]

        # Mass and cost scaling factors
        self.myfloat['bulkhead_mass_factor']     = 1.0     # Scaling for unaccounted bulkhead mass
        self.myfloat['ring_mass_factor']         = 1.0     # Scaling for unaccounted stiffener mass
        self.myfloat['shell_mass_factor']        = 1.0     # Scaling for unaccounted shell mass
        self.myfloat['column_mass_factor']       = 1.0    # Scaling for unaccounted column mass
        self.myfloat['outfitting_mass_fraction'] = 0.0    # Fraction of additional outfitting mass for each column
        self.myfloat['ballast_cost_rate']        = 0.1   # Cost factor for ballast mass [$/kg]
        self.myfloat['material_cost_rate']       = 1.1  # Cost factor for column mass [$/kg]
        self.myfloat['labor_cost_rate']          = 1.0  # Cost factor for labor time [$/min]
        self.myfloat['painting_cost_rate']       = 14.4  # Cost factor for column surface finishing [$/m^2]
        self.myfloat['outfitting_cost_rate']     = 1.5*1.1  # Cost factor for outfitting mass [$/kg]
        self.myfloat['mooring_cost_factor']        = 1.1     # Cost factor for mooring mass [$/kg]

        # Safety factors
        self.myfloat['gamma_f'] = 1.0 # Safety factor on loads
        self.myfloat['gamma_b'] = 1.0  # Safety factor on buckling
        self.myfloat['gamma_m'] = 1.0  # Safety factor on materials
        self.myfloat['gamma_n'] = 1.0  # Safety factor on consequence of failure
        self.myfloat['gamma_fatigue'] = 1.0 # Not used

        # Properties of rotor-nacelle-assembly (RNA)
        self.myfloat['rna_mass']   = 350e3 # Mass [kg]
        self.myfloat['rna_I']      = 1e5*np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        self.myfloat['rna_cg']     = np.zeros(3)
        self.myfloat['rna_force']  = np.zeros(3)
        self.myfloat['rna_moment'] = np.zeros(3)

        # Mooring constraints
        self.myfloat['max_offset'] = 0.1*self.myfloat['water_depth'] # Max surge/sway offset [m]      
        self.myfloat['max_survival_heel']   = 10.0 # Max heel (pitching) angle [deg]
        self.myfloat['operational_heel']   = 5.0 # Max heel (pitching) angle [deg]

        # Design constraints
        self.myfloat['max_draft'] = 200.0                # For manufacturability of rolling steel
        self.myfloat['max_taper_ratio'] = 0.4                # For manufacturability of rolling steel
        self.myfloat['min_diameter_thickness_ratio'] = 120.0 # For weld-ability

        # API 2U flag
        self.myfloat['loading'] = 'axial' #'hydrostatic'

        # Porperties of turbine tower
        self.myfloat['hub_height']              = 77.6                              # Length from tower main to top (not including freeboard) [m]
        self.myfloat['tower_section_height']    = 77.6/nSecTow * np.ones(nSecTow) # Length of each tower section [m]
        self.myfloat['tower_outer_diameter']    = np.linspace(6.5, 3.87, nSecTow+1) # Diameter at each tower section node (linear lofting between) [m]
        self.myfloat['tower_wall_thickness']    = np.linspace(0.027, 0.019, nSecTow) # Diameter at each tower section node (linear lofting between) [m]
        self.myfloat['tower_buckling_length']   = 30.0                              # Tower buckling reinforcement spacing [m]
        self.myfloat['tower_outfitting_factor'] = 1.07                              # Scaling for unaccounted tower mass in outfitting

        
        
    def testMassPropertiesSpar(self):
        # Column geometry
        self.myfloat['main.permanent_ballast_height'] = 0.0 # Height above keel for permanent ballast [m]
        self.myfloat['main.freeboard']                = 10.0 # Height extension above waterline [m]
        self.myfloat['main.section_height'] = np.array([49.0, 59.0, 8.0, 14.0])  # Length of each section [m]
        self.myfloat['main.outer_diameter'] = np.array([9.4, 9.4, 9.4, 6.5, 6.5]) # Diameter at each section node (linear lofting between) [m]
        self.myfloat['main.wall_thickness'] = 0.05 * np.ones(nSec)               # Shell thickness at each section node (linear lofting between) [m]
        self.myfloat['main.bulkhead_thickness'] = 0.05*np.array([1, 1, 0, 1, 0]) # Locations/thickness of internal bulkheads at section interfaces [m]
        self.myfloat['main.buoyancy_tank_diameter'] = 0.0
        self.myfloat['main.buoyancy_tank_height'] = 0.0

        # Column ring stiffener parameters
        self.myfloat['main.stiffener_web_height']       = 0.10 * np.ones(nSec) # (by section) [m]
        self.myfloat['main.stiffener_web_thickness']    = 0.04 * np.ones(nSec) # (by section) [m]
        self.myfloat['main.stiffener_flange_width']     = 0.10 * np.ones(nSec) # (by section) [m]
        self.myfloat['main.stiffener_flange_thickness'] = 0.02 * np.ones(nSec) # (by section) [m]
        self.myfloat['main.stiffener_spacing']          = np.array([1.5, 2.8, 3.0, 5.0]) # (by section) [m]

        # Mooring parameters
        self.myfloat['number_of_mooring_connections']    = 3             # Evenly spaced around structure
        self.myfloat['mooring_lines_per_connection']    = 1             # Evenly spaced around structure
        self.myfloat['mooring_type']               = 'chain'       # Options are chain, nylon, polyester, fiber, or iwrc
        self.myfloat['anchor_type']                = 'suctionpile' # Options are SUCTIONPILE or DRAGEMBEDMENT
        self.myfloat['mooring_diameter']           = 0.09          # Diameter of mooring line/chain [m]
        self.myfloat['fairlead_location']          = 0.384615 # Want 70.0          # Distance below waterline for attachment [m]
        self.myfloat['fairlead_offset_from_shell'] = 0.5           # Offset from shell surface for mooring attachment [m]
        self.myfloat['mooring_line_length']        = 300+902.2         # Unstretched mooring line length
        self.myfloat['anchor_radius']              = 853.87        # Distance from centerline to sea floor landing [m]
        self.myfloat['fairlead_support_outer_diameter'] = 3.2    # Diameter of all fairlead support elements [m]
        self.myfloat['fairlead_support_wall_thickness'] = 0.0175 # Thickness of all fairlead support elements [m]

        # Other variables to avoid divide by zeros, even though it won't matter
        self.myfloat['radius_to_offset_column'] = 15.0
        self.myfloat['off.section_height'] = 1.0 * np.ones(nSec)
        self.myfloat['off.outer_diameter'] = 5.0 * np.ones(nSec+1)
        self.myfloat['off.wall_thickness'] = 0.1 * np.ones(nSec)
        self.myfloat['off.permanent_ballast_height'] = 0.1
        self.myfloat['off.stiffener_web_height'] = 0.1 * np.ones(nSec)
        self.myfloat['off.stiffener_web_thickness'] =  0.1 * np.ones(nSec)
        self.myfloat['off.stiffener_flange_width'] =  0.1 * np.ones(nSec)
        self.myfloat['off.stiffener_flange_thickness'] =  0.1 * np.ones(nSec)
        self.myfloat['off.stiffener_spacing'] =  0.1 * np.ones(nSec)
        self.myfloat['off.freeboard'] =  0.1
        self.myfloat['pontoon_outer_diameter'] = 1.0
        self.myfloat['pontoon_wall_thickness'] = 0.1

        # Set environment to that used in OC3 testing campaign
        self.myfloat['water_depth'] = 320.0  # Distance to sea floor [m]
        self.myfloat['significant_wave_height']        = 0.0    # Significant wave height [m]
        self.myfloat['significant_wave_period']           = 1e3    # Wave period [s]
        self.myfloat['wind_reference_speed']        = 0.0    # Wind reference speed [m/s]
        self.myfloat['wind_reference_height']        = 119.0  # Wind reference height [m]

        self.myfloat.run_model()

        m_top = np.pi*3.2**2.0*0.05*7850.0
        ansys_m_bulk  = 13204.0 + 2.0*27239.0 + m_top
        ansys_m_shell = 80150.0 + 32060.0 + 79701.0 + 1251800.0
        ansys_m_stiff = 1390.9*52 + 1282.2 + 1121.2 + 951.44*3
        ansys_m_spar = ansys_m_bulk + ansys_m_shell + ansys_m_stiff
        ansys_cg     = -58.926
        ansys_Ixx    = 2178400000.0 + m_top*(0.25*3.2**2.0 + (10-ansys_cg)**2)
        ansys_Iyy    = 2178400000.0 + m_top*(0.25*3.2**2.0 + (10-ansys_cg)**2)
        ansys_Izz    = 32297000.0 + 0.5*m_top*3.2**2.0
        ansys_I      = np.array([ansys_Ixx, ansys_Iyy, ansys_Izz, 0.0, 0.0, 0.0])

        npt.assert_allclose(ansys_m_bulk, self.myfloat['main.bulkhead_mass'].sum(), rtol=0.03) # ANSYS uses R_od, we use R_id, top cover seems unaccounted for
        npt.assert_allclose(ansys_m_shell, self.myfloat['main.cyl_mass.mass'].sum(), rtol=0.01)
        npt.assert_allclose(ansys_m_stiff, self.myfloat['main.stiffener_mass'].sum(), rtol=0.01)
        npt.assert_allclose(ansys_m_spar, self.myfloat['main.column_total_mass'].sum(), rtol=0.01)
        npt.assert_allclose(ansys_cg, self.myfloat['main.z_center_of_mass'], rtol=0.01)
        npt.assert_allclose(ansys_I, self.myfloat['main.I_column'], rtol=0.02)

    def testSemiRuns(self):

        # SEMI SPECIFIC OPTIONS
        # Add in offset columns and truss elements
        self.myfloat['number_of_offset_columns'] = 3
        self.myfloat['cross_attachment_pontoons_int']   = 1 # Lower-Upper main-to-offset connecting cross braces
        self.myfloat['lower_attachment_pontoons_int']   = 1 # Lower main-to-offset connecting pontoons
        self.myfloat['upper_attachment_pontoons_int']   = 1 # Upper main-to-offset connecting pontoons
        self.myfloat['lower_ring_pontoons_int']         = 1 # Lower ring of pontoons connecting offset columns
        self.myfloat['upper_ring_pontoons_int']         = 1 # Upper ring of pontoons connecting offset columns
        self.myfloat['outer_cross_pontoons_int']        = 1 # Auxiliary ring connecting V-cross braces

        # Set environment to that used in OC4 testing campaign
        self.myfloat['water_depth']             = 200.0  # Distance to sea floor [m]
        self.myfloat['significant_wave_height'] = 10.8   # Significant wave height [m]
        self.myfloat['significant_wave_period'] = 9.8    # Wave period [s]
        self.myfloat['wind_reference_speed']    = 11.0   # Wind reference speed [m/s]
        self.myfloat['wind_reference_height']   = 119.0  # Wind reference height [m]

        # Column geometry
        self.myfloat['main.permanent_ballast_height'] = 10.0 # Height above keel for permanent ballast [m]
        self.myfloat['main.freeboard']                = 10.0 # Height extension above waterline [m]
        self.myfloat['main.section_height'] = np.array([49.0, 59.0, 8.0, 14.0])  # Length of each section [m]
        self.myfloat['main.outer_diameter'] = np.array([9.4, 9.4, 9.4, 6.5, 6.5]) # Diameter at each section node (linear lofting between) [m]
        self.myfloat['main.wall_thickness'] = 0.05 * np.ones(nSec)               # Shell thickness at each section node (linear lofting between) [m]
        self.myfloat['main.bulkhead_thickness'] = 0.05*np.array([1, 1, 0, 1, 0]) # Locations/thickness of internal bulkheads at section interfaces [m]

        # Auxiliary column geometry
        self.myfloat['radius_to_offset_column']         = 33.333 * np.cos(np.pi/6) # Centerline of main column to centerline of offset column [m]
        self.myfloat['off.permanent_ballast_height'] = 0.1                      # Height above keel for permanent ballast [m]
        self.myfloat['off.freeboard']                = 12.0                     # Height extension above waterline [m]
        self.myfloat['off.section_height']           = np.array([6.0, 0.1, 15.9, 10]) # Length of each section [m]
        self.myfloat['off.outer_diameter']           = np.array([24, 24, 12, 12, 12]) # Diameter at each section node (linear lofting between) [m]
        self.myfloat['off.wall_thickness']           = 0.06 * np.ones(nSec)         # Shell thickness at each section node (linear lofting between) [m]

        # Column ring stiffener parameters
        self.myfloat['main.stiffener_web_height']       = 0.10 * np.ones(nSec) # (by section) [m]
        self.myfloat['main.stiffener_web_thickness']    = 0.04 * np.ones(nSec) # (by section) [m]
        self.myfloat['main.stiffener_flange_width']     = 0.10 * np.ones(nSec) # (by section) [m]
        self.myfloat['main.stiffener_flange_thickness'] = 0.02 * np.ones(nSec) # (by section) [m]
        self.myfloat['main.stiffener_spacing']          = 0.40 * np.ones(nSec) # (by section) [m]

        # Auxiliary column ring stiffener parameters
        self.myfloat['off.stiffener_web_height']       = 0.10 * np.ones(nSec) # (by section) [m]
        self.myfloat['off.stiffener_web_thickness']    = 0.04 * np.ones(nSec) # (by section) [m]
        self.myfloat['off.stiffener_flange_width']     = 0.01 * np.ones(nSec) # (by section) [m]
        self.myfloat['off.stiffener_flange_thickness'] = 0.02 * np.ones(nSec) # (by section) [m]
        self.myfloat['off.stiffener_spacing']          = 0.40 * np.ones(nSec) # (by section) [m]

        # Pontoon parameters
        self.myfloat['pontoon_outer_diameter']    = 3.2    # Diameter of all pontoon/truss elements [m]
        self.myfloat['pontoon_wall_thickness']    = 0.0175 # Thickness of all pontoon/truss elements [m]
        self.myfloat['main_pontoon_attach_lower'] = -20.0  # Lower z-coordinate on main where truss attaches [m]
        self.myfloat['main_pontoon_attach_upper'] = 10.0   # Upper z-coordinate on main where truss attaches [m]

        # Mooring parameters
        self.myfloat['number_of_mooring_connections']    = 3             # Evenly spaced around structure
        self.myfloat['mooring_lines_per_connection']    = 1             # Evenly spaced around structure
        self.myfloat['mooring_type']               = 'chain'       # Options are chain, nylon, polyester, fiber, or iwrc
        self.myfloat['anchor_type']                = 'suctionpile' # Options are SUCTIONPILE or DRAGEMBEDMENT
        self.myfloat['mooring_diameter']           = 0.0766        # Diameter of mooring line/chain [m]
        self.myfloat['fairlead']                   = 14.0          # Distance below waterline for attachment [m]
        self.myfloat['fairlead_offset_from_shell'] = 0.5           # Offset from shell surface for mooring attachment [m]
        self.myfloat['mooring_line_length']        = 835.5+300         # Unstretched mooring line length
        self.myfloat['anchor_radius']              = 837.6+300.0         # Distance from centerline to sea floor landing [m]
        self.myfloat['fairlead_support_outer_diameter'] = 3.2    # Diameter of all fairlead support elements [m]
        self.myfloat['fairlead_support_wall_thickness'] = 0.0175 # Thickness of all fairlead support elements [m]
        
        self.myfloat.run_model()
        self.assertTrue(True)
        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestRegression))
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())
    
    
